"""
String Constants used in Hops-Util: Environment variables, Kafka Config, SSL Config etc.
"""


class HTTP_CONFIG:
    """
    HTTP String constants
    """
    HTTP_CONTENT_TYPE = "Content-type"
    HTTP_APPLICATION_JSON = "application/json"
    HTTP_AUTHORIZATION = "Authorization"
    HTTP_POST = "POST"
    HTTP_PUT = "PUT"
    HTTP_GET = "GET"
    HTTP_DELETE = "DELETE"
    HTTP_UNAUTHORIZED = 401


class ENV_VARIABLES:
    """
    Environment variable names (accessible in os.environ)
    """
    KAFKA_BROKERS_ENV_VAR = "KAFKA_BROKERS"
    ELASTIC_ENDPOINT_ENV_VAR = "ELASTIC_ENDPOINT"
    PWD_ENV_VAR = "PWD"
    KAFKA_VERSION_ENV_VAR = "KAFKA_VERSION"
    LIVY_VERSION_ENV_VAR = "LIVY_VERSION"
    SPARK_VERSION_ENV_VAR = "SPARK_VERSION"
    REST_ENDPOINT_END_VAR = "REST_ENDPOINT"
    TENSORFLOW_VERSION_ENV_VAR = "TENSORFLOW_VERSION"
    CUDA_VERSION_ENV_VAR = "CUDA_VERSION"
    HOPSWORKS_VERSION_ENV_VAR = "HOPSWORKS_VERSION"
    HADOOP_VERSION_ENV_VAR = "HADOOP_VERSION"
    HADOOP_USER_NAME_ENV_VAR = "HADOOP_USER_NAME"
    HADOOP_HOME = "HADOOP_HOME"
    HADOOP_CLASSPATH_GLOB = "HADOOP_CLASSPATH_GLOB"
    HDFS_USER_ENV_VAR = "HDFS_USER"
    HOPSWORKS_USER_ENV_VAR = "HOPSWORKS_USER"
    PATH_ENV_VAR = "PATH"
    PYTHONPATH_ENV_VAR = "PYTHONPATH"
    JOB_NAME_ENV_VAR = "HOPSWORKS_JOB_NAME"
    KERNEL_ID_ENV_VAR = "HOPSWORKS_KERNEL_ID"
    HOPSWORKS_PROJECT_ID_ENV_VAR = "HOPSWORKS_PROJECT_ID"
    HOPSWORKS_PROJECT_NAME_ENV_VAR = "HOPSWORKS_PROJECT_NAME"
    API_KEY_ENV_VAR = "API_KEY"
    SERVING_API_KEY_ENV_VAR = "SERVING_API_KEY"
    REGION_NAME_ENV_VAR = "REGION_NAME"
    FLINK_CONF_DIR = "FLINK_CONF_DIR"
    FLINK_LIB_DIR = "FLINK_LIB_DIR"
    REQUESTS_VERIFY = "REQUESTS_VERIFY"
    REQUESTS_VERIFY_ENV_VAR = "REQUESTS_VERIFY"
    DOMAIN_CA_TRUSTSTORE_ENV_VAR = "DOMAIN_CA_TRUSTSTORE"
    DOMAIN_CA_TRUSTSTORE_PEM_ENV_VAR = "DOMAIN_CA_TRUSTSTORE_PEM"
    SECRETS_DIR_ENV_VAR = "SECRETS_DIR"
    SPARK_IS_DRIVER = "IS_HOPS_DRIVER"


class KAFKA_SSL_CONFIG:
    """
    Kafka SSL constant strings for configuration
    """
    SSL = "SSL"
    SSL_TRUSTSTORE_LOCATION_CONFIG = "ssl.truststore.location"
    SSL_TRUSTSTORE_LOCATION_DOC = "The location of the trust store file. "
    SSL_TRUSTSTORE_PASSWORD_CONFIG = "ssl.truststore.password"
    SSL_TRUSTSTORE_PASSWORD_DOC = "The password for the trust store file. If a password is not set access to the truststore is still available, but integrity checking is disabled."
    SSL_KEYSTORE_LOCATION_CONFIG = "ssl.keystore.location"
    SSL_KEYSTORE_PASSWORD_CONFIG = "ssl.keystore.password"
    SSL_KEY_PASSWORD_CONFIG = "ssl.key.password"
    SECURITY_PROTOCOL_CONFIG = "security.protocol"
    SSL_CERTIFICATE_LOCATION_CONFIG = "ssl.certificate.location"
    SSL_CA_LOCATION_CONFIG = "ssl.ca.location"
    SSL_PRIVATE_KEY_LOCATION_CONFIG = "ssl.key.location"
    SSL_ENDPOINT_IDENTIFICATION_ALGORITHM_CONFIG = "ssl.endpoint.identification.algorithm"


# General SSL config properties

class SSL_CONFIG:
    """
    General SSL configuration constants for Hops-TLS
    """
    KEYSTORE_SUFFIX = "__kstore.jks"
    TRUSTSTORE_SUFFIX = "__tstore.jks"
    PASSWORD_SUFFIX = "__cert.key"

    K_CERTIFICATE_CONFIG = "k_certificate"
    T_CERTIFICATE_CONFIG = "t_certificate"
    PEM_CLIENT_CERTIFICATE_CONFIG = "client.pem"
    PEM_CLIENT_KEY_CONFIG = "client_key.pem"
    PEM_CA_CHAIN_CERTIFICATE_CONFIG = "ca_chain.pem"
    DOMAIN_CA_TRUSTSTORE = "domain_ca_truststore"
    CRYPTO_MATERIAL_PASSWORD = "material_passwd"
    PEM_CA_ROOT_CERT = "/srv/hops/kagent/host-certs/hops_root_ca.pem"
    SSL_ENABLED = "ipc.server.ssl.enabled"


class KAFKA_PRODUCER_CONFIG:
    """
    Constant strings for Kafka producers
    """
    BOOTSTRAP_SERVERS_CONFIG = "bootstrap.servers"
    KEY_SERIALIZER_CLASS_CONFIG = "key.serializer"
    VALUE_SERIALIZER_CLASS_CONFIG = "value.serializer"


class KAFKA_CONSUMER_CONFIG:
    """
    Constant strings for Kafka consumers
    """
    GROUP_ID_CONFIG = "group.id"
    ENABLE_AUTO_COMMIT_CONFIG = "enable.auto.commit"
    AUTO_COMMIT_INTERVAL_MS_CONFIG = "auto.commit.interval.ms"
    SESSION_TIMEOUT_MS_CONFIG = "session.timeout.ms"
    KEY_DESERIALIZER_CLASS_CONFIG = "key.deserializer"
    VALUE_DESERIALIZER_CLASS_CONFIG = "value.deserializer"
    AUTO_OFFSET_RESET_CONFIG = "auto.offset.reset"
    ENABLE_AUTO_COMMIT_CONFIG = "enable.auto.commit"
    KEY_DESERIALIZER_CLASS_CONFIG = "key.deserializer"
    VALUE_DESERIALIZER_CLASS_CONFIG = "value.deserializer"


class SPARK_CONFIG:
    """
    Spark string constants
    """
    SPARK_SCHEMA_FIELD_METADATA = "metadata"
    SPARK_SCHEMA_FIELDS = "fields"
    SPARK_SCHEMA_FIELD_NAME = "name"
    SPARK_SCHEMA_FIELD_TYPE = "type"
    SPARK_SCHEMA_ELEMENT_TYPE = "elementType"
    SPARK_OVERWRITE_MODE = "overwrite"
    SPARK_APPEND_MODE = "append"
    SPARK_WRITE_DELIMITER = "delimiter"
    SPARK_INFER_SCHEMA = "inferSchema"
    SPARK_WRITE_HEADER = "header"
    SPARK_TF_CONNECTOR_RECORD_TYPE = "recordType"
    SPARK_TF_CONNECTOR_RECORD_TYPE_EXAMPLE = "Example"
    SPARK_LONG_TYPE = "long"
    SPARK_SHORT_TYPE = "short"
    SPARK_BYTE_TYPE = "byte"
    SPARK_INTEGER_TYPE = "integer"
    SPARK_INT_TYPE = "int"
    SPARK_FLOAT_TYPE = "float"
    SPARK_DOUBLE_TYPE = 'double'
    SPARK_DECIMAL_TYPE = "decimal"
    SPARK_BIGINT_TYPE = "bigint"
    SPARK_SMALLINT_TYPE = "smallint"
    SPARK_STRING_TYPE = "string"
    SPARK_BINARY_TYPE = "binary"
    SPARK_NUMERIC_TYPES = [SPARK_BIGINT_TYPE,
                           SPARK_DECIMAL_TYPE,
                           SPARK_INTEGER_TYPE,
                           SPARK_INT_TYPE,
                           SPARK_DOUBLE_TYPE,
                           SPARK_LONG_TYPE,
                           SPARK_FLOAT_TYPE,
                           SPARK_SHORT_TYPE]
    SPARK_STRUCT = "struct"
    SPARK_ARRAY = "array"
    SPARK_ARRAY_DOUBLE = "array<double>"
    SPARK_ARRAY_INTEGER = "array<integer>"
    SPARK_ARRAY_INT = "array<int>"
    SPARK_ARRAY_BIGINT = "array<bigint>"
    SPARK_ARRAY_FLOAT = "array<float>"
    SPARK_ARRAY_DECIMAL = "array<decimal>"
    SPARK_ARRAY_STRING = "array<string>"
    SPARK_ARRAY_LONG = "array<long>"
    SPARK_ARRAY_BINARY = "array<binary>"
    SPARK_VECTOR = "vector"
    SPARK_SQL_CATALOG_IMPLEMENTATION = "spark.sql.catalogImplementation"
    SPARK_SQL_CATALOG_HIVE = "hive"
    SPARK_JDBC_FORMAT = "jdbc"
    SPARK_JDBC_URL = "url"
    SPARK_JDBC_DBTABLE = "dbtable"
    SPARK_JDBC_USER = "user"
    SPARK_JDBC_PW = "password"


class MODEL_SERVING:
    MODELS_DATASET = "Models"
    MODEL_SERVER_TENSORFLOW_SERVING = "TENSORFLOW_SERVING"
    MODEL_SERVER_PYTHON = "PYTHON"
    MODEL_SERVERS = [MODEL_SERVER_TENSORFLOW_SERVING, MODEL_SERVER_PYTHON]
    SERVING_TOOL_DEFAULT = "DEFAULT"
    SERVING_TOOL_KFSERVING = "KFSERVING"
    SERVING_TOOLS = [SERVING_TOOL_DEFAULT, SERVING_TOOL_KFSERVING]
    INFERENCE_LOGGING_MODEL_INPUTS = "MODEL_INPUTS"
    INFERENCE_LOGGING_PREDICTIONS = "PREDICTIONS"
    INFERENCE_LOGGING_ALL = "ALL"
    INFERENCE_LOGGING_MODES = [INFERENCE_LOGGING_MODEL_INPUTS, INFERENCE_LOGGING_PREDICTIONS, INFERENCE_LOGGING_ALL]
    SERVING_ACTION_START = "START"
    SERVING_ACTION_STOP = "STOP"
    SERVING_ACTIONS = [SERVING_ACTION_STOP, SERVING_ACTION_STOP]
    SERVING_START_OR_STOP_PATH_PARAM = "?action="


class FEATURE_STORE:
    """
     Featurestore constants
    """
    TRAINING_DATASET_PROVENANCE_FEATUREGROUP = "featuregroup"
    TRAINING_DATASET_PROVENANCE_VERSION = "version"
    MAX_CORRELATION_MATRIX_COLUMNS = 50
    TRAINING_DATASET_CSV_FORMAT = "csv"
    TRAINING_DATASET_TSV_FORMAT = "tsv"
    TRAINING_DATASET_PARQUET_FORMAT = "parquet"
    TRAINING_DATASET_TFRECORDS_FORMAT = "tfrecords"
    TRAINING_DATASET_TFRECORD_FORMAT = "tfrecord"
    TRAINING_DATASET_AVRO_FORMAT = "avro"
    TRAINING_DATASET_ORC_FORMAT = "orc"
    TRAINING_DATASET_NPY_FORMAT = "npy"
    TRAINING_DATASET_IMAGE_FORMAT = "image"
    TRAINING_DATASET_HDF5_FORMAT = "hdf5"
    TRAINING_DATASET_PETASTORM_FORMAT = "petastorm"
    TRAINING_DATASET_NPY_SUFFIX = ".npy"
    TRAINING_DATASET_HDF5_SUFFIX = ".hdf5"
    TRAINING_DATASET_CSV_SUFFIX = ".csv"
    TRAINING_DATASET_TSV_SUFFIX = ".tsv"
    TRAINING_DATASET_PARQUET_SUFFIX = ".parquet"
    TRAINING_DATASET_AVRO_SUFFIX = ".avro"
    TRAINING_DATASET_ORC_SUFFIX = ".orc"
    TRAINING_DATASET_IMAGE_SUFFIX = ".image"
    TRAINING_DATASET_TFRECORDS_SUFFIX = ".tfrecords"
    TRAINING_DATASET_PETASTORM_SUFFIX = ".petastorm"
    TRAINING_DATASET_SUPPORTED_FORMATS = [
        TRAINING_DATASET_TSV_FORMAT,
        TRAINING_DATASET_CSV_FORMAT,
        TRAINING_DATASET_PARQUET_FORMAT,
        TRAINING_DATASET_TFRECORDS_FORMAT,
        TRAINING_DATASET_TFRECORD_FORMAT,
        TRAINING_DATASET_NPY_FORMAT,
        TRAINING_DATASET_HDF5_FORMAT,
        TRAINING_DATASET_AVRO_FORMAT,
        TRAINING_DATASET_ORC_FORMAT,
        TRAINING_DATASET_IMAGE_FORMAT,
        TRAINING_DATASET_PETASTORM_FORMAT
    ]
    FEATURE_GROUP_INSERT_APPEND_MODE = "append"
    FEATURE_GROUP_INSERT_OVERWRITE_MODE = "overwrite"
    FEATURESTORE_SUFFIX = "_featurestore"
    TRAINING_DATASETS_SUFFIX = "_Training_Datasets"
    TRAINING_DATASET_TF_RECORD_SCHEMA_FILE_NAME = "tf_record_schema.txt"
    TF_RECORD_SCHEMA_FEATURE = "feature"
    TF_RECORD_SCHEMA_FEATURE_FIXED = "fixed_len"
    TF_RECORD_SCHEMA_FEATURE_VAR = "var_len"
    TF_RECORD_SCHEMA_TYPE = "type"
    TF_RECORD_SCHEMA_SHAPE = "shape"
    TF_RECORD_INT_TYPE = "int"
    TF_RECORD_FLOAT_TYPE = "float"
    TF_RECORD_STRING_TYPE = "string"
    TF_RECORD_INT_ARRAY_SPARK_TYPES = [SPARK_CONFIG.SPARK_ARRAY_INTEGER, SPARK_CONFIG.SPARK_ARRAY_BIGINT,
                                       SPARK_CONFIG.SPARK_ARRAY_INT, SPARK_CONFIG.SPARK_ARRAY_LONG]
    TF_RECORD_INT_SPARK_TYPES = [SPARK_CONFIG.SPARK_INTEGER_TYPE, SPARK_CONFIG.SPARK_BIGINT_TYPE,
                                 SPARK_CONFIG.SPARK_INT_TYPE, SPARK_CONFIG.SPARK_LONG_TYPE]
    TF_RECORD_STRING_SPARK_TYPES = [SPARK_CONFIG.SPARK_STRING_TYPE, SPARK_CONFIG.SPARK_BINARY_TYPE]
    TF_RECORD_STRING_ARRAY_SPARK_TYPES = [SPARK_CONFIG.SPARK_ARRAY_STRING, SPARK_CONFIG.SPARK_ARRAY_BINARY]
    TF_RECORD_FLOAT_SPARK_TYPES = [SPARK_CONFIG.SPARK_FLOAT_TYPE, SPARK_CONFIG.SPARK_DECIMAL_TYPE,
                                   SPARK_CONFIG.SPARK_DOUBLE_TYPE]
    TF_RECORD_FLOAT_ARRAY_SPARK_TYPES = [SPARK_CONFIG.SPARK_ARRAY_FLOAT, SPARK_CONFIG.SPARK_ARRAY_DECIMAL,
                                         SPARK_CONFIG.SPARK_ARRAY_DOUBLE, SPARK_CONFIG.SPARK_VECTOR]
    RECOGNIZED_TF_RECORD_TYPES = [SPARK_CONFIG.SPARK_VECTOR, SPARK_CONFIG.SPARK_ARRAY_BINARY,
                                  SPARK_CONFIG.SPARK_ARRAY_STRING, SPARK_CONFIG.SPARK_ARRAY_DECIMAL,
                                  SPARK_CONFIG.SPARK_ARRAY_DOUBLE, SPARK_CONFIG.SPARK_ARRAY_FLOAT,
                                  SPARK_CONFIG.SPARK_ARRAY_LONG, SPARK_CONFIG.SPARK_ARRAY_INTEGER,
                                  SPARK_CONFIG.SPARK_BINARY_TYPE, SPARK_CONFIG.SPARK_STRING_TYPE,
                                  SPARK_CONFIG.SPARK_DECIMAL_TYPE, SPARK_CONFIG.SPARK_DOUBLE_TYPE,
                                  SPARK_CONFIG.SPARK_FLOAT_TYPE, SPARK_CONFIG.SPARK_LONG_TYPE,
                                  SPARK_CONFIG.SPARK_INT_TYPE, SPARK_CONFIG.SPARK_INTEGER_TYPE,
                                  SPARK_CONFIG.SPARK_ARRAY_BIGINT, SPARK_CONFIG.SPARK_BIGINT_TYPE,
                                  SPARK_CONFIG.SPARK_ARRAY_INT]
    DATAFRAME_TYPE_SPARK = "spark"
    DATAFRAME_TYPE_NUMPY = "numpy"
    DATAFRAME_TYPE_PYTHON = "python"
    DATAFRAME_TYPE_PANDAS = "pandas"
    JDBC_TRUSTSTORE_ARG = "sslTrustStore"
    JDBC_TRUSTSTORE_PW_ARG = "trustStorePassword"
    JDBC_KEYSTORE_ARG = "sslKeyStore"
    JDBC_KEYSTORE_PW_ARG = "keyStorePassword"
    IMPORT_HOPS_UTIL_FEATURESTORE_HELPER = "import io.hops.util.featurestore.FeaturestoreHelper"


class PETASTORM_CONFIG:
    """
    Petastorm String constants
    """
    FILESYSTEM_FACTORY = "pyarrow_filesystem"
    SCHEMA = "schema"
    LIBHDFS = "libhdfs"


class MYSQL_CONFIG:
    """ MYSQL string constants """
    MYSQL_DATA_TYPES = [
        "None", "INT(11)", "TINYINT(1)", "SMALLINT(5)", "MEDIUMINT(7)", "BIGINT(20)", "FLOAT", "DOUBLE", "DECIMAL",
        "DATE", "DATETIME", "TIMESTAMP", "TIME", "YEAR", "CHAR", "VARCHAR(25)", "VARCHAR(125)", "VARCHAR(225)",
        "VARCHAR(500)", "VARCHAR(1000)", "VARCHAR(2000)", "VARCHAR(5000)", "VARCHAR(10000)", "BLOB", "TEXT",
        "TINYBLOB", "TINYTEXT", "MEDIUMBLOB", "MEDIUMTEXT", "LONGBLOB", "LONGTEXT", "JSON"
    ]
    MYSQL_BIGINT_TYPE = "BIGINT(20)"
    MYSQL_SMALLINT_TYPE = "SMALLINT(5)"
    MYSQL_CHAR_TYPE = "CHAR"
    MYSQL_INTEGER_TYPE = "INT(11)"
    MYSQL_VARCHAR_1000_TYPE = "VARCHAR(1000)"
    MYSQL_BLOB_TYPE = "BLOB"


class HIVE_CONFIG:
    """
    Hive string constants
    """
    HIVE_DATA_TYPES = [
        "TINYINT", "SMALLINT", "INT", "BIGINT", "FLOAT", "DOUBLE",
        "DECIMAL", "TIMESTAMP", "DATE", "INTERVAL", "STRING", "VARCHAR",
        "CHAR", "BOOLEAN", "BINARY", "ARRAY", "MAP", "STRUCT", "UNIONTYPE"
    ]
    HIVE_BIGINT_TYPE = "BIGINT"
    HIVE_INT_TYPE = "INT"
    HIVE_CHAR_TYPE = "CHAR"


class REST_CONFIG:
    """
    REST endpoints and JSON properties used for communicating with Hopsworks REST API
    """
    JSON_KEYSTOREPWD = "keyStorePwd"
    JSON_SCHEMA_CONTENTS = "contents"

    JSON_TYPE = "type"

    JSON_FEATURESTORE_UPDATE_STATS_QUERY_PARAM = "updateStats"
    JSON_FEATURESTORE_UPDATE_METADATA_QUERY_PARAM = "updateMetadata"
    JSON_FEATURESTORE_UPDATE_JOB_QUERY_PARAM = "updateJob"
    JSON_FEATURESTORE_ENABLE_ONLINE_QUERY_PARAM = "enableOnline"
    JSON_FEATURESTORE_DISABLE_ONLINE_QUERY_PARAM = "disableOnline"
    JSON_FEATURESTORE_UPDATE_STATISTICS_SETTINGS = "updateStatsSettings"
    JSON_FEATURESTORE_SETTINGS_ENTITY_NAME_MAX_LENGTH = "featurestoreEntityNameMaxLength"
    JSON_FEATURESTORE_SETTINGS_ENTITY_DESCRIPTION_MAX_LENGTH = "featurestoreEntityDescriptionMaxLength"
    JSON_FEATURESTORE_SETTINGS_CACHED_FEATUREGROUP_DTO_TYPE = "cachedFeaturegroupDtoType"
    JSON_FEATURESTORE_SETTINGS_EXTERNAL_TRAINING_DATASET_TYPE = "externalTrainingDatasetType"
    JSON_FEATURESTORE_SETTINGS_FEATURESTORE_REGEX = "featurestoreRegex"
    JSON_FEATURESTORE_SETTINGS_HOPSFS_CONNECTOR_DTO_TYPE = "hopsfsConnectorDtoType"
    JSON_FEATURESTORE_SETTINGS_HOPSFS_CONNECTOR_TYPE = "hopsfsConnectorType"
    JSON_FEATURESTORE_SETTINGS_HOPSFS_TRAINING_DATASET_TYPE = "hopsfsTrainingDatasetType"
    JSON_FEATURESTORE_SETTINGS_JDBC_CONNECTOR_DTO_TYPE = "jdbcConnectorDtoType"
    JSON_FEATURESTORE_SETTINGS_JDBC_CONNECTOR_TYPE = "jdbcConnectorType"
    JSON_FEATURESTORE_SETTINGS_JDBC_CONNECTOR_ARGUMENTS_MAX_LEN = "jdbcStorageConnectorArgumentsMaxLength"
    JSON_FEATURESTORE_SETTINGS_JDBC_CONNECTOR_CONNECTION_STRING_MAX_LEN = "jdbcStorageConnectorConnectionstringMaxLength"
    JSON_FEATURESTORE_SETTINGS_ON_DEMAND_FEATUREGROUP_DTO_TYPE = "onDemandFeaturegroupDtoType"
    JSON_FEATURESTORE_SETTINGS_ON_DEMAND_FEATUREGROUP_SQL_QUERY_MAX_LEN = "onDemandFeaturegroupSqlQueryMaxLength"
    JSON_FEATURESTORE_SETTINGS_S3_CONNECTOR_DTO_TYPE = "s3ConnectorDtoType"
    JSON_FEATURESTORE_SETTINGS_S3_CONNECTOR_TYPE = "s3ConnectorType"
    JSON_FEATURESTORE_SETTINGS_S3_CONNECTOR_ACCESS_KEY_MAX_LEN = "s3StorageConnectorAccesskeyMaxLength"
    JSON_FEATURESTORE_SETTINGS_S3_CONNECTOR_BUCKET_MAX_LEN = "s3StorageConnectorBucketMaxLength"
    JSON_FEATURESTORE_SETTINGS_S3_CONNECTOR_SECRET_KEY_MAX_LEN = "s3StorageConnectorSecretkeyMaxLength"
    JSON_FEATURESTORE_SETTINGS_STORAGE_CONNECTOR_DESCRIPTION_MAX_LEN = "storageConnectorDescriptionMaxLength"
    JSON_FEATURESTORE_SETTINGS_STORAGE_CONNECTOR_NAME_MAX_LEN = "storageConnectorDescriptionMaxLength"
    JSON_FEATURESTORE_SETTINGS_HIVE_SUGGESTED_FEATURE_TYPES = "suggestedHiveFeatureTypes"
    JSON_FEATURESTORE_SETTINGS_MYSQL_SUGGESTED_FEATURE_TYPES = "suggestedMysqlFeatureTypes"
    JSON_FEATURESTORE_SETTINGS_TRAINING_DATASET_DATA_FORMATS = "trainingDatasetDataFormats"
    JSON_FEATURESTORE_SETTINGS_TRAINING_DATASET_TYPE = "trainingDatasetType"
    JSON_FEATURESTORE_SETTINGS = "settings"
    JSON_FEATURESTORE_STORAGE_CONNECTORS = "storageConnectors"
    JSON_FEATURESTORE_SETTINGS_IMPORT_CONNECTORS = "featureImportConnectors"
    JSON_FEATURESTORE_SETTINGS_ONLINE_ENABLED = "onlineFeaturestoreEnabled"

    JSON_FEATURESTORE_JOB_FEATUREGROUP_ID = "featuregroupId"
    JSON_FEATURESTORE_JOB_TRAINING_DATASET_ID = "trainingDatasetId"
    JSON_FEATURESTORE_JOB_LAST_COMPUTED = "lastComputed"
    JSON_FEATURESTORE_JOB_STATUS = "jobStatus"
    JSON_FEATURESTORE_JOB_NAME = "jobName"
    JSON_FEATURESTORE_JOB_ID = "jobId"
    JSON_FEATURESTORE_LOCATION = "location"

    JSON_FEATUREGROUP_ON_DEMAND_QUERY = "query"
    JSON_FEATUREGROUP_JDBC_CONNECTOR_NAME = "jdbcConnectorName"
    JSON_FEATUREGROUP_JDBC_CONNECTOR_ID = "jdbcConnectorId"
    JSON_FEATUREGROUP_TYPE = "type"
    JSON_FEATUREGROUP_NAME = "name"
    JSON_FEATUREGROUP_ID = "id"
    JSON_FEATUREGROUP_VERSION = "version"
    JSON_FEATUREGROUP_JOBS = "jobs"
    JSON_FEATUREGROUP_FEATURES = "features"
    JSON_FEATUREGROUP_DESCRIPTION = "description"
    JSON_FEATUREGROUP_CREATED = "created"
    JSON_FEATUREGROUP_CREATOR = "creator"
    JSON_FEATUREGROUPS = "featuregroups"
    JSON_FEATUREGROUP_ONLINE = "onlineEnabled"
    JSON_FEATUREGROUP_HUDI = "hudiEnabled"
    JSON_FEATUREGROUP_FEATURE_HISTOGRAM_ENABLED = "featHistEnabled"
    JSON_FEATUREGROUP_FEATURE_CORRELATION_ENABLED = "featCorrEnabled"
    JSON_FEATUREGROUP_DESCRIPTIVE_STATISTICS_ENABLED = "descStatsEnabled"
    JSON_FEATUREGROUP_STATISTIC_COLUMNS = "statisticColumns"

    JSON_ONLINE_FEATUREGROUP_ID = "id"
    JSON_ONLINE_FEATUREGROUP_DB = "dbName"
    JSON_ONLINE_FEATUREGROUP_TABLE = "tableName"
    JSON_ONLINE_FEATUREGROUP_TABLE_TYPE = "tableType"
    JSON_ONLINE_FEATUREGROUP_TABLE_ROWS = "tableRows"
    JSON_ONLINE_FEATUREGROUP_SIZE = "size"

    JSON_FEATURESTORE = "featurestore"
    JSON_FEATURESTORE_ID = "featurestoreId"
    JSON_FEATURESTORE_NAME = "featurestoreName"
    JSON_FEATURESTORE_PROJECT_ID = "projectId"
    JSON_FEATURESTORE_PROJECT_NAME = "projectName"
    JSON_FEATURESTORE_INODE_ID = "inodeId"
    JSON_FEATURESTORE_DESCRIPTION = "featurestoreDescription"
    JSON_FEATURESTORE_HDFS_PATH = "hdfsStorePath"
    JSON_FEATURESTORE_ONLINE_CONNECTOR = "onlineFeaturestoreConnector"
    JSON_FEATURESTORE_ONLINE_ENABLED = "onlineEnabled"
    JSON_FEATURESTORE_ONLINE_FEATURESTORE_TYPE = "onlineFeaturestoreType"
    JSON_FEATURESTORE_OFFLINE_FEATURESTORE_TYPE = "offlineFeaturestoreType"
    JSON_FEATURESTORE_ONLINE_FEATURESTORE_NAME = "onlineFeaturestoreName"
    JSON_FEATURESTORE_OFFLINE_FEATURESTORE_NAME = "offlineFeaturestoreName"

    JSON_FEATURE_NAME = "name"
    JSON_FEATURE_TYPE = "type"
    JSON_FEATURE_INDEX = "index"
    JSON_FEATURE_DESCRIPTION = "description"
    JSON_FEATURE_PRIMARY = "primary"
    JSON_FEATURE_PARTITION = "partition"
    JSON_FEATURE_ONLINE_TYPE = "onlineType"
    JSON_FEATURE_FEATUREGROUP = "featuregroup"
    JSON_FEATURE_VERSION = "version"

    JSON_TRAINING_DATASET_EXTERNAL_TYPE = "EXTERNAL_TRAINING_DATASET"
    JSON_TRAINING_DATASET_HOPSFS_TYPE = "HOPSFS_TRAINING_DATASET"
    JSON_TRAINING_DATASET_TYPE = "trainingDatasetType"
    JSON_TRAINING_DATASET_CONNECTOR_NAME = "storageConnectorName"
    JSON_TRAINING_DATASET_CONNECTOR_ID = "storageConnectorId"
    JSON_TRAINING_DATASET_SIZE = "size"
    JSON_TRAINING_DATASET_ID = "id"
    JSON_TRAINING_DATASET_NAME = "name"
    JSON_TRAINING_DATASETS = "trainingDatasets"
    JSON_TRAINING_DATASET_HDFS_STORE_PATH = "hdfsStorePath"
    JSON_TRAINING_DATASET_FORMAT = "dataFormat"
    JSON_TRAINING_DATASET_SCHEMA = "features"
    JSON_TRAINING_DATASET_VERSION = "version"
    JSON_TRAINING_DATASET_CREATOR = "creator"
    JSON_TRAINING_DATASET_CREATED = "created"
    JSON_TRAINING_DATASET_DESCRIPTION = "description"
    JSON_TRAINING_DATASET_JOBNAME = "jobName"
    JSON_TRAINING_DATASET_INODE_ID = "inodeId"
    JSON_TRAINING_DATASET_FEATURES = "features"
    JSON_TRAINING_DATASET_JOBS = "jobs"

    JSON_FEATURESTORE_HOPSFS_CONNECTOR_HOPSFS_PATH = "hopsfsPath"
    JSON_FEATURESTORE_HOPSFS_CONNECTOR_DATASET_NAME = "datasetName"

    JSON_FEATURESTORE_JDBC_CONNECTOR_CONNECTION_STRING = "connectionString"
    JSON_FEATURESTORE_JDBC_CONNECTOR_ARGUMENTS = "arguments"

    JSON_FEATURESTORE_S3_ACCESS_KEY = "accessKey"
    JSON_FEATURESTORE_S3_SECRET_KEY = "secretKey"
    JSON_FEATURESTORE_S3_BUCKET = "bucket"

    JSON_FEATURESTORE_CONNECTOR_NAME = "name"
    JSON_FEATURESTORE_CONNECTOR_DESCRIPTION = "description"
    JSON_FEATURESTORE_CONNECTOR_ID = "id"
    JSON_FEATURESTORE_CONNECTOR_FEATURESTORE_ID = "featurestoreId"
    JSON_FEATURESTORE_CONNECTOR_TYPE = "storageConnectorType"

    JSON_SCHEMA_VERSION = "version"
    JSON_KEYSTORE = "keyStore"

    HOPSWORKS_REST_RESOURCE = "hopsworks-api/api"
    HOPSWORKS_SCHEMA_RESOURCE = "schema"
    HOPSWORKS_FEATURESTORES_RESOURCE = "featurestores"
    HOPSWORKS_FEATURESTORE_METADATA_RESOURCE = "metadata"
    HOPSWORKS_FEATUREGROUPS_RESOURCE = "featuregroups"
    HOPSWORKS_TRAININGDATASETS_RESOURCE = "trainingdatasets"
    HOPSWORKS_FEATUREGROUP_CLEAR_RESOURCE = "clear"
    HOPSWORKS_FEATUREGROUPS_SYNC_RESOURCE = "sync"
    HOPSWORKS_SERVING_RESOURCE = "serving"
    HOPSWORKS_INFERENCE_RESOURCE = "inference"
    HOPSWORKS_MODEL_REGISTRY_RESOURCE = "modelregistries"
    HOPSWORKS_MODELS_RESOURCE = "models"
    HOPSWORKS_USERS_RESOURCE = "users"
    HOPSWORKS_ADMIN_RESOURCE = "admin"
    HOPSWORKS_FEATURESTORES_STORAGE_CONNECTORS_RESOURCE = "storageconnectors"
    HOPSWORKS_ONLINE_FEATURESTORE_STORAGE_CONNECTOR_RESOURCE = "onlinefeaturestore"
    HOPSWORKS_FEATURESTORE_TAGS_RESOURCE = "tags"
    HOPSWORKS_VARIABLES_RESOURCE = "variables"
    HOPSWORKS_ENDPOINT = "hopsworks_endpoint"

    HOPSWORKS_EXPERIMENTS_RESOURCE = "experiments"

    HOPSWORKS_KAFKA_RESOURCE = "kafka"
    HOPSWORKS_TOPICS_RESOURCE = "topics"
    HOPSWORKS_SUBJECTS_RESOURCE = "subjects"

    HOPSWORKS_AS_SHARED = "asShared"
    HOPSWORKS_SHARED = "shared"
    HOPSWORKS_PROJECT_RESOURCE = "project"
    HOPSWORKS_USER_RESOURCE = "users"
    HOPSWORKS_PROJECT_INFO_RESOURCE = "getProjectInfo"
    HOPSWORKS_JOBS_RESOURCE = "jobs"
    HOPSWORKS_SECRETS_RESOURCE = "secrets"
    HOPSWORKS_EXECUTIONS_RESOURCE = "executions"
    HOPSWORKS_DATASETS_RESOURCE = "dataset"
    HOPSWORKS_PROJECT_CREDENTIALS_RESOURCE = "credentials"
    HOPSWORKS_PROJECT_CLIENT = "client"
    HOPSWORKS_AUTH_RESOURCE = "auth"
    HOPSWORKS_AUTH_RESOURCE_REGISTER = "register"

    HOPSWORKS_XATTR_RESOURCE = "xattrs"
    HOPSWORKS_ELASTIC_RESOURCE = "elastic"
    HOPSWORKS_ELASTIC_JWT_RESOURCE = "jwt"

    JSON_ERROR_CODE = "errorCode"
    JSON_ERROR_MSG = "errorMsg"
    JSON_USR_MSG = "usrMsg"

    JWT_TOKEN = "token.jwt"

    JSON_SERVING_ID = "id"
    JSON_SERVING_NAME = "name"
    JSON_SERVING_MODEL_PATH = "modelPath"
    JSON_SERVING_MODEL_NAME = "modelName"
    JSON_SERVING_MODEL_VERSION = "modelVersion"
    JSON_SERVING_MODEL_SERVER = "modelServer"
    JSON_SERVING_TOOL = "servingTool"
    JSON_SERVING_ARTIFACT_VERSION = "artifactVersion"
    JSON_SERVING_PREDICTOR = "predictor"
    JSON_SERVING_TRANSFORMER = "transformer"
    JSON_SERVING_REQUESTED_INSTANCES = "requestedInstances"
    JSON_SERVING_PREDICTOR_RESOURCE_CONFIG = "predictorResourceConfig"
    JSON_SERVING_REQUESTED_TRANSFORMER_INSTANCES = "requestedTransformerInstances"
    JSON_SERVING_BATCHING_ENABLED = "batchingEnabled"
    JSON_SERVING_KAFKA_TOPIC_DTO = "kafkaTopicDTO"
    JSON_SERVING_CREATE_KAFKA_TOPIC = "CREATE"
    JSON_SERVING_DONT_CREATE_KAFKA_TOPIC = "NONE"
    JSON_SERVING_INFERENCE_LOGGING = "inferenceLogging"
    JSON_SERVING_CREATED = "created"
    JSON_SERVING_CREATOR = "creator"
    JSON_SERVING_REVISION = "revision"
    JSON_SERVING_STATUS = "status"
    JSON_SERVING_AVAILABLE_INSTANCES = "availableInstances"
    JSON_SERVING_AVAILABLE_TRANSFORMER_INSTANCES = "availableTransformerInstances"
    JSON_SERVING_EXTERNAL_IP = "externalIP"
    JSON_SERVING_EXTERNAL_PORT = "externalPort"
    JSON_SERVING_INTERNAL_IPS = "internalIPs"
    JSON_SERVING_INTERNAL_PORT = "internalPort"
    JSON_SERVING_INTERNAL_PATH = "internalPath"

    JSON_KAFKA_TOPIC_SCHEMA_VERSION = "schemaVersion"
    JSON_KAFKA_TOPIC_NAME = "name"
    JSON_KAFKA_NUM_PARTITIONS = "numOfPartitions"
    JSON_KAFKA_NUM_REPLICAS = "numOfReplicas"

    HOPSWORKS_CLOUD_RESOURCE = "cloud"
    HOPSWORKS_AWS_CLOUD_SESSION_TOKEN_RESOURCE = "aws/session-token"
    HOPSWORKS_CLOUD_ROLE_MAPPINGS_RESOURCE = "role-mappings"
    HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_ROLE = "roleARN"
    HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_SESSION = "roleSessionName"
    HOPSWORKS_CLOUD_SESSION_TOKEN_RESOURCE_QUERY_SESSION_DURATION = "durationSeconds"

    JSON_ACCESS_KEY_ID = "accessKeyId"
    JSON_SECRET_KEY_ID = "secretAccessKey"
    JSON_SESSION_TOKEN_ID = "sessionToken"
    JSON_ARRAY_ITEMS = "items"
    JSON_CLOUD_ROLE = "cloudRole"


class DELIMITERS:
    """
    String delimiters constants
    """
    SLASH_DELIMITER = "/"
    COMMA_DELIMITER = ","
    TAB_DELIMITER = "\t"
    COLON_DELIMITER = ":"
    DOT_DELIMITER = "."
    AMPERSAND_DELIMITER = "&"
    SEMI_COLON_DELIMITER = ";"
    JDBC_CONNECTION_STRING_VALUE_DELIMITER = "="
    JDBC_CONNECTION_STRING_DELIMITER = ";"
    QUESTION_MARK_DELIMITER = "?"


class S3_CONFIG:
    """
    String constants for S3
    """
    S3_FILE_PREFIX = "s3a://"
    S3_ACCESS_KEY_ENV = "fs.s3a.access.key"
    S3_SECRET_KEY_ENV = "fs.s3a.secret.key"
    S3_SESSION_KEY_ENV = "fs.s3a.session.token"
    S3_CREDENTIAL_PROVIDER_ENV = "fs.s3a.aws.credentials.provider"
    S3_TEMPORARY_CREDENTIAL_PROVIDER = "org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider"
    S3_TRAINING_DATASETS_FOLDER = "TRAINING_DATASETS"
    AWS_ACCESS_KEY_ID_ENV = "AWS_ACCESS_KEY_ID"
    AWS_SECRET_ACCESS_KEY_ENV = "AWS_SECRET_ACCESS_KEY"
    AWS_SESSION_TOKEN_ENV = "AWS_SESSION_TOKEN"


class AWS:
    DEFAULT_REGION = 'default'
    SECRETS_MANAGER = "secretsmanager"
    PARAMETER_STORE = "parameterstore"


class LOCAL:
    LOCAL_STORE = "local"


class XATTRS:
    XATTRS_PARAM_NAME = 'name'


class ELASTICSEARCH_CONFIG:
    SSL_CONFIG = "es.net.ssl"
    NODES_WAN_ONLY = "es.nodes.wan.only"
    NODES = "es.nodes"
    SSL_KEYSTORE_LOCATION = "es.net.ssl.keystore.location"
    SSL_KEYSTORE_PASSWORD = "es.net.ssl.keystore.pass"
    SSL_TRUSTSTORE_LOCATION = "es.net.ssl.truststore.location"
    SSL_TRUSTSTORE_PASSWORD = "es.net.ssl.truststore.pass"
    HTTP_AUTHORIZATION = "es.net.http.header.Authorization"
    INDEX = "es.resource"


"""

Utility functions to manage jobs in Hopsworks.

"""
from hops import constants, util, hdfs
from hops.exceptions import RestAPIError
import json


def create_job(name, job_config):
    """
    Create a job in Hopsworks

    Args:
        name: Name of the job to be created.
        job_config: A dictionary representing the job configuration

    Returns:
        HTTP(S)Connection
    """
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    job_config["appName"] = name
    method = constants.HTTP_CONFIG.HTTP_PUT
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_JOBS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   name
    response = util.send_request(method, resource_url, data=json.dumps(job_config), headers=headers)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not create job (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return response_object


def _job_execution_action(name, args=None):
    """
    Manages execution for the given job, start or stop. Submits an http request to the HOPSWORKS REST API.

    Returns:
        The job status.
    """
    method = constants.HTTP_CONFIG.HTTP_PUT
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_JOBS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   name + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_EXECUTIONS_RESOURCE

    response = util.send_request(method, resource_url, args)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not perform action on job's execution (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return response_object


def start_job(name, args=None):
    """
    Start an execution of the job. Only one execution can be active for a job.

    Returns:
        The job status.
    """
    method = constants.HTTP_CONFIG.HTTP_POST
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_JOBS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   name + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_EXECUTIONS_RESOURCE

    response = util.send_request(method, resource_url, args)
    response_object = response.json()
    if response.status_code >= 400:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not perform action on job's execution (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return response_object


def stop_job(name):
    """
    Stop the current execution of the job.
    Returns:
        The job status.
    """
    method = constants.HTTP_CONFIG.HTTP_PUT
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_JOBS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   name + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_EXECUTIONS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   "{EXECUTION_ID}" + constants.DELIMITERS.SLASH_DELIMITER + \
                   "status"

    status = {"status": "stopped"}
    # If no execution_id was provided, stop all active executions
    # Get all active execution IDs
    executions = get_executions(name,
                                "?filter_by=state:INITIALIZING,RUNNING,ACCEPTED,NEW,NEW_SAVING,SUBMITTED,"
                                "STARTING_APP_MASTER")
    responses = []
    if executions['count'] > 0:
        for execution in executions['items']:
            responses.append(util.http(resource_url.replace("{EXECUTION_ID}", str(execution['id'])), headers, method,
                                       json.dumps(status)))
    return responses


def get_executions(name, query=""):
    """
    Get a list of the currently running executions for this job.
    Returns:
        The job status.
    """
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_JOBS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   name + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_EXECUTIONS_RESOURCE + query
    response = util.send_request(method, resource_url)
    response_object = response.json()
    if response.status_code >= 500:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not get current job's execution (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))
    if response.status_code >= 400:
        return None
    return response_object



"""
A module for setting up Kafka Brokers and Consumers on the Hops platform. It hides the complexity of
configuring Kafka by providing utility methods such as:

    - `get_broker_endpoints()`.
    - `get_security_protocol()`.
    - `get_kafka_default_config()`.
    - etc.

Using these utility functions you can setup Kafka with the Kafka client-library of your choice, e.g SparkStreaming or
confluent-kafka-python. For example, assuming that you have created a topic called "test" on Hopsworks and that you
have installed confluent-kafka-python inside your project's anaconda environment:

    >>> from hops import kafka
    >>> from confluent_kafka import Producer, Consumer
    >>> TOPIC_NAME = "test"
    >>> config = kafka.get_kafka_default_config()
    >>> producer = Producer(config)
    >>> consumer = Consumer(config)
    >>> consumer.subscribe(["test"])
    >>> # wait a little while before executing the rest of the code (put it in a different Jupyter cell)
    >>> # so that the consumer get chance to subscribe (asynchronous call)
    >>> for i in range(0, 10):
    >>> producer.produce(TOPIC_NAME, "message {}".format(i), "key", callback=delivery_callback)
    >>> # Trigger the sending of all messages to the brokers, 10sec timeout
    >>> producer.flush(10)
    >>> for i in range(0, 10):
    >>> msg = consumer.poll(timeout=5.0)
    >>> if msg is not None:
    >>>     print('Consumed Message: {} from topic: {}'.format(msg.value(), msg.topic()))
    >>> else:
    >>>     print("Topic empty, timeout when trying to consume message")


Similarly, you can define a pyspark kafka consumer as follows, using the spark session defined in variable `spark`

    >>> from hops import kafka
    >>> from hops import tls
    >>> TOPIC_NAME = "test"
    >>> df = spark \.format("kafka")
    >>> .option("kafka.bootstrap.servers", kafka.get_broker_endpoints())
    >>> .option("kafka.ssl.truststore.location", tls.get_trust_store())
    >>> .option("kafka.ssl.truststore.password", tls.get_key_store_pwd())
    >>> .option("kafka.ssl.keystore.location", tls.get_key_store())
    >>> .option("kafka.ssl.keystore.password", tls.get_key_store_pwd())
    >>> .option("kafka.ssl.key.password", tls.get_trust_store_pwd())
    >>> .option("subscribe", TOPIC_NAME)
    >>> .load()
"""

import os
from hops import constants, tls, util, hdfs
from hops.exceptions import RestAPIError
import json
import sys

# for backwards compatibility
try:
    from ast import literal_eval
    from io import BytesIO
    from avro.io import DatumReader, BinaryDecoder
    import avro.schema
except:
    pass

def get_broker_endpoints():
    """
    Get Kafka broker endpoints as a string with broker-endpoints "," separated

    Returns:
        a string with broker endpoints comma-separated
    """
    return os.environ[constants.ENV_VARIABLES.KAFKA_BROKERS_ENV_VAR].replace("INTERNAL://","")


def get_security_protocol():
    """
    Gets the security protocol used for communicating with Kafka brokers in a Hopsworks cluster

    Returns:
        the security protocol for communicating with Kafka brokers in a Hopsworks cluster
    """
    return constants.KAFKA_SSL_CONFIG.SSL


def get_broker_endpoints_list():
    """
    Get Kafka broker endpoints as a list

    Returns:
        a list with broker endpoint strings
    """
    return get_broker_endpoints().split(",")


def get_kafka_default_config():
    """
    Gets a default configuration for running secure Kafka on Hops

    Returns:
         dict with config_property --> value
    """
    default_config = {
        constants.KAFKA_PRODUCER_CONFIG.BOOTSTRAP_SERVERS_CONFIG: get_broker_endpoints(),
        constants.KAFKA_SSL_CONFIG.SECURITY_PROTOCOL_CONFIG: get_security_protocol(),
        constants.KAFKA_SSL_CONFIG.SSL_CA_LOCATION_CONFIG: tls.get_ca_chain_location(),
        constants.KAFKA_SSL_CONFIG.SSL_CERTIFICATE_LOCATION_CONFIG: tls.get_client_certificate_location(),
        constants.KAFKA_SSL_CONFIG.SSL_PRIVATE_KEY_LOCATION_CONFIG: tls.get_client_key_location(),
        "group.id": "something"
    }
    return default_config


def get_schema(topic):
    """
    Gets the Avro schema for a particular Kafka topic.

    Args:
        :topic: Kafka topic name

    Returns:
        Avro schema as a string object in JSON format
    """
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_KAFKA_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_TOPICS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   topic + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_SUBJECTS_RESOURCE
    response = util.send_request(method, resource_url)
    response_object = response.json()

    if response.status_code != 200:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise RestAPIError("Could not get Avro schema (url: {}), server response: \n "
                           "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, user msg: {}".format(
            resource_url, response.status_code, response.reason, error_code, error_msg, user_msg))

    return response_object['schema']


def parse_avro_msg(msg, avro_schema):
    """
    Parses an avro record using a specified avro schema

    Args:
        :msg: the avro message to parse
        :avro_schema: the avro schema

    Returns:
         The parsed/decoded message
    """
    reader = DatumReader(avro_schema)
    message_bytes = BytesIO(msg)
    decoder = BinaryDecoder(message_bytes)
    return reader.read(decoder)


def convert_json_schema_to_avro(json_schema):
    """
    Parses a JSON kafka topic schema returned by Hopsworks REST API into an avro schema

    Args:
       :json_schema: the json schema to convert

    Returns:
         the avro schema
    """

    return avro.schema.parse(json_schema)



class KafkaTopicDTO(object):
    """
    Represents a KafkaTopic in Hopsworks
    """

    def __init__(self, kafka_topic_dto_json):
        """
        Initialize the kafka topic from JSON payload returned by Hopsworks REST API

        Args:
            :kafka_topic_dto_json: JSON data about the kafka topic returned from Hopsworks REST API
        """
        self.name = kafka_topic_dto_json[constants.REST_CONFIG.JSON_KAFKA_TOPIC_NAME]
        self.schema_version = kafka_topic_dto_json[constants.REST_CONFIG.JSON_KAFKA_TOPIC_SCHEMA_VERSION]



"""

Model module is used for exporting, versioning, attaching metadata to models. In pipelines it should be used to query the Model Repository for the best model version to use during inference.

"""

from hops import constants, util, hdfs, project, dataset
from hops.experiment_impl.util import experiment_utils
from hops.exceptions import RestAPIError
import json
import sys
import os
import time
import six
import glob
from six import string_types

import logging
log = logging.getLogger(__name__)

import urllib3
urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)

import warnings

warnings.warn(
    "The `model` module was deprecated with the "
    "introduction of the new `HSML` client library to interact with the Hopsworks "
    "Model Registry. It is recommended to use the new hsml library as this model module will be removed in a future release. "
    "Please see: https://docs.hopsworks.ai/machine-learning-api/latest/",
    DeprecationWarning,
)

def get_best_model(name, metric, direction):
    """
    Get the best model version by sorting on attached metadata such as model accuracy.

    For example if you run this:

    >>> from hops import model
    >>> from hops.model import Metric
    >>> model.get_best_model('mnist', 'accuracy', Metric.MAX)

    It will return the mnist version where the 'accuracy' is the highest.

    Args:
        :name: name of the model
        :metric: name of the metric to compare
        :direction: whether metric should be maximized or minimized to find the best model

    Returns:
        The best model

    Raises:
        :ModelNotFound: if the model was not found
    """

    if direction.upper() == Metric.MAX:
        direction = "desc"
    elif direction.upper() == Metric.MIN:
        direction = "asc"
    else:
        raise Exception("Invalid direction, should be Metric.MAX or Metric.MIN")

    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}

    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_MODEL_REGISTRY_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_MODELS_RESOURCE + \
                   "?filter_by=name_eq:" + name + "&sort_by=" + metric + ":" + direction + "&limit=1"

    response_object = util.send_request('GET', resource_url, headers=headers)

    if not response_object.ok or 'items' not in json.loads(response_object.content.decode("UTF-8")):
        raise ModelNotFound("No model with name {} and metric {} could be found.".format(name, metric))

    return json.loads(response_object.content.decode("UTF-8"))['items'][0]


def get_model(name, version, project_name=None):
    """
    Get a specific model version given a model name and a version.

    For example if you run this:

    >>> from hops import model
    >>> model.get_model('mnist', 1)

    You will get version 1 of the model 'mnist'

    Args:
        :name: name of the model
        :version: version of the model
        :project_name name of the project parent of the model. By default, this project is the current project running
        the experiment

    Returns:
        The specified model version

    Raises:
        :ModelNotFound: if the model was not found
    """

    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    resource_url = constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_MODEL_REGISTRY_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   str(experiment_utils.get_model_registry_id(project_name)) + constants.DELIMITERS.SLASH_DELIMITER + \
                   constants.REST_CONFIG.HOPSWORKS_MODELS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER + \
                   str(name) + "_" + str(version)

    response_object = util.send_request('GET', resource_url, headers=headers)

    if response_object.ok:
        return response_object

    raise ModelNotFound("No model with name: {} and version {} could be found".format(name, version))


def download_model(name, version=None, project_name=None, overwrite=False):
    """
    Download from the Hopsworks Models dataset an archive (zip file) containing the model artifacts.
    You first need to use the project.connect function to connect to Hopsworks.
    If the Models dataset where the model resides is a shared dataset from another project,
    then you need to specify the name of the project that owns the Models dataset was shared from.

    For example if you run this:

    >>> from hops import model
    >>> # If connecting from an external client, you need to connect to Hopsworks
    >>> project.connect(...) # see project module for documentation
    >>> model.download_model('mnist')

    Args:
        :name: name of the model
        :version: version of the model. If omitted, all versions of the model will be included in the archive.
        :project_name name of the project parent of the model. By default, this project is the current project running
        the experiment
        :overwrite: Whether to overwrite the model archive  file if it already exists

    Returns:
        A zip file containing the model artifacts

    Raises:
        :ModelArchiveExists: if the model archive that contains the model artifacts already exists
    """
    if project_name is None:
        project_name = hdfs.project_name()

    # Check if model archive already exists and if it should be deleted, otherwise return an error
    model_dir = '/Projects/' + project_name + "/Models/" + name
    if version is not None:
        model_dir += "/" + str(version)
        name += str(version)
    archive_path = model_dir + ".zip"
    name += ".zip"
    if dataset.path_exists(archive_path):
        if overwrite:
            dataset.delete(archive_path, block=True)
        else:
            raise ModelArchiveExists("Model archive file already exists at {}. Either set overwrite=True or remove the file manually.".format(archive_path))

    log.debug("Preparing the model archive...")
    dataset.compress(model_dir, block=True, project_name=project_name)
    log.debug("Downloading the model archive...")
    dataset.download(archive_path, file = name)


def export(model_path, model_name, model_version=None, overwrite=False, metrics=None, description=None,
           synchronous=True, synchronous_timeout=120, project=None):
    """
    Copies a trained model to the Models directory in the project and creates the directory structure of:

    >>> Models
    >>>      |
    >>>      - model_name
    >>>                 |
    >>>                 - version_x
    >>>                 |
    >>>                 - version_y

    For example if you run this:

    >>> from hops import model
    >>> model.export("iris_knn.pkl", "irisFlowerClassifier", metrics={'accuracy': accuracy})

    It will copy the local model file "iris_knn.pkl" to /Projects/projectname/Models/irisFlowerClassifier/1/model.pkl
    on HDFS, and overwrite in case there already exists a file with the same name in the directory.

    If "model" is a directory on the local path exported by TensorFlow, and you run:

    >>> model.export("/model", "mnist", metrics={'accuracy': accuracy, 'loss': loss})

    It will copy the model directory contents to /Projects/projectname/Models/mnist/1/ , e.g the "model.pb" file and
    the "variables" directory.

    Args:
        :model_path: absolute path to the trained model (HDFS or local)
        :model_name: name of the model
        :model_version: version of the model
        :overwrite: boolean flag whether to overwrite in case a model already exists in the exported directory
        :metrics: dict of evaluation metrics to attach to model
        :description: description about the model
        :synchronous: whether to synchronously wait for the model to be indexed in the models rest endpoint
        :synchronous_timeout: max timeout in seconds for waiting for the model to be indexed
        :project: the name of the project where the model should be saved to (default: current project). Note, the project must share its 'Models' dataset and make it writeable for this client.

    Returns:
        The path to where the model was exported

    Raises:
        :ValueError: if there was an error with the model due to invalid user input
        :ModelNotFound: if the model was not found
        :MultipleModelFilesFound: if multiple model files are found
    """

    # Make sure model name is a string, users could supply numbers
    model_name = str(model_name)

    if not isinstance(model_path, string_types):
        model_path = model_path.decode()

    if not description:
        description = 'A collection of models for ' + model_name

    project_path = hdfs.project_path(project)

    assert hdfs.exists(project_path + "Models"), "Your project is missing a dataset named Models, please create it."

    if not hdfs.exists(model_path) and not os.path.exists(model_path):
        raise ValueError("the provided model_path: {} , does not exist in HDFS or on the local filesystem".format(
            model_path))

    _check_model_file(model_path)

    # make sure metrics are numbers
    if metrics:
        _validate_metadata(metrics)

    model_dir_hdfs = project_path + constants.MODEL_SERVING.MODELS_DATASET + \
                     constants.DELIMITERS.SLASH_DELIMITER + model_name + constants.DELIMITERS.SLASH_DELIMITER

    if not hdfs.exists(model_dir_hdfs):
        hdfs.mkdir(model_dir_hdfs)
        hdfs.chmod(model_dir_hdfs, "ug+rwx")

    # User did not specify model_version, pick the current highest version + 1, set to 1 if no model exists
    version_list = []
    if not model_version and hdfs.exists(model_dir_hdfs):
        model_version_directories = hdfs.ls(model_dir_hdfs)
        for version_dir in model_version_directories:
            try:
                if hdfs.isdir(version_dir):
                    version_list.append(int(version_dir[len(model_dir_hdfs):]))
            except:
                pass
        if len(version_list) > 0:
            model_version = max(version_list) + 1

    if not model_version:
        model_version = 1

    # Path to directory in HDFS to put the model files
    model_version_dir_hdfs = model_dir_hdfs + str(model_version)

    # If version directory already exists and we are not overwriting it then fail
    if not overwrite and hdfs.exists(model_version_dir_hdfs):
        raise ValueError("Could not create model directory: {}, the path already exists, "
                         "set flag overwrite=True "
                         "to remove the version directory and create the correct directory structure".format(model_version_dir_hdfs))

    # Overwrite version directory by deleting all content (this is needed for Provenance to register Model as deleted)
    if overwrite and hdfs.exists(model_version_dir_hdfs):
       hdfs.delete(model_version_dir_hdfs, recursive=True)
       hdfs.mkdir(model_version_dir_hdfs)

    # At this point we can create the version directory if it does not exist
    if not hdfs.exists(model_version_dir_hdfs):
       hdfs.mkdir(model_version_dir_hdfs)

    # Export the model files
    if os.path.exists(model_path):
        export_dir=_export_local_model(model_path, model_version_dir_hdfs, overwrite)
    else:
        export_dir=_export_hdfs_model(model_path, model_version_dir_hdfs, overwrite)

    # Attach modelName_modelVersion to experiment directory
    if project is None:
        model_project_name = hdfs.project_name()
    else :
        model_project_name = project
    experiment_project_name = hdfs.project_name()
    model_summary = { 'name': model_name, 'projectName': model_project_name, 'version': model_version, 'metrics':  metrics,
                      'experimentId': None, 'experimentProjectName': experiment_project_name,
                      'description':  description}
    if 'ML_ID' in os.environ:
        model_summary['experimentId'] = os.environ['ML_ID']
        # Attach link from experiment to model
        experiment_json = experiment_utils._populate_experiment_model(model_name + '_' + str(model_version), project=project)
        experiment_utils._attach_experiment_xattr(os.environ['ML_ID'], experiment_json, 'MODEL_UPDATE')

    # Attach model metadata to models version folder
    experiment_utils._attach_model_xattr(model_name + "_" + str(model_version), experiment_utils.dumps(model_summary), model_project_name)

    # Model metadata is attached asynchronously by Epipe, therefore this necessary to ensure following steps in a pipeline will not fail
    if synchronous:
        start_time = time.time()
        sleep_seconds = 5
        for i in range(int(synchronous_timeout/sleep_seconds)):
            try:
                time.sleep(sleep_seconds)
                log.info("Polling " + model_name + " version " + str(model_version) + " for model availability.")
                resp = get_model(model_name, model_version, project_name=project)
                if resp.ok:
                    log.info("Exported model " + model_name + " as version " + str(model_version) + " successfully.")
                    return
                log.warn(model_name + " not ready yet, retrying in " + str(sleep_seconds) + " seconds.")
            except ModelNotFound:
                pass
        log.warn("Model not available during polling, set a higher value for synchronous_timeout to wait longer.")

    return export_dir

def _check_model_file(model_path):
    """Checks the existance of a single model file

     Args:
        :model_path: the path to model files

     Returns:
        The model file path
     Raises:
        :MultipleModelFilesFound: if multiple model files are found
    """

    MODEL_FILE_EXTENSIONS = [".pb", ".joblib", ".pkl", ".pickle"]

    files = None
    model_files_path = os.path.join(model_path, "*")
    if os.path.exists(model_path):  # local
        if os.path.isdir(model_path):
            files = glob.iglob(model_files_path)
    else:
        if hdfs.isdir(model_path):  # hdfs
            files = hdfs.glob(model_files_path)

    if files is None:
        return

    model_file = None
    for path in files:
        ext = os.path.splitext(path)[1]
        if ext in MODEL_FILE_EXTENSIONS:
            if model_file is not None:
                raise MultipleModelFilesFound("Multiple model files found in the model path")
            model_file = os.path.split(path)[1]
    return model_file

def _export_local_model(local_model_path, model_dir_hdfs, overwrite):
    """
    Exports a local directory of model files to Hopsworks "Models" dataset

     Args:
        :local_model_path: the path to the local model files
        :model_dir_hdfs: path to the directory in HDFS to put the model files
        :overwrite: boolean flag whether to overwrite existing model files

    Returns:
           the path to the exported model files in HDFS
    """
    if os.path.isdir(local_model_path):
        if not local_model_path.endswith(constants.DELIMITERS.SLASH_DELIMITER):
            local_model_path = local_model_path + constants.DELIMITERS.SLASH_DELIMITER
        for filename in os.listdir(local_model_path):
            hdfs.copy_to_hdfs(local_model_path + filename, model_dir_hdfs, overwrite=overwrite)
    elif os.path.isfile(local_model_path):
        hdfs.copy_to_hdfs(local_model_path, model_dir_hdfs, overwrite=overwrite)

    return model_dir_hdfs

def _export_hdfs_model(hdfs_model_path, model_dir_hdfs, overwrite):
    """
    Exports a hdfs directory of model files to Hopsworks "Models" dataset

     Args:
        :hdfs_model_path: the path to the model files in hdfs
        :model_dir_hdfs: path to the directory in HDFS to put the model files
        :overwrite: boolean flag whether to overwrite in case a model already exists in the exported directory

    Returns:
           the path to the exported model files in HDFS
    """
    if hdfs.isdir(hdfs_model_path):
        for file_source_path in hdfs.ls(hdfs_model_path):
            model_name = file_source_path
            if constants.DELIMITERS.SLASH_DELIMITER in file_source_path:
                last_index = model_name.rfind(constants.DELIMITERS.SLASH_DELIMITER)
                model_name = model_name[last_index + 1:]
            dest_path = model_dir_hdfs + constants.DELIMITERS.SLASH_DELIMITER + model_name
            hdfs.cp(file_source_path, dest_path, overwrite=overwrite)
    elif hdfs.isfile(hdfs_model_path):
        model_name = hdfs_model_path
        if constants.DELIMITERS.SLASH_DELIMITER in hdfs_model_path:
            last_index = model_name.rfind(constants.DELIMITERS.SLASH_DELIMITER)
            model_name = model_name[last_index + 1:]
        dest_path = model_dir_hdfs + constants.DELIMITERS.SLASH_DELIMITER + model_name
        hdfs.cp(hdfs_model_path, dest_path, overwrite=overwrite)

    return model_dir_hdfs

def _validate_metadata(metrics):
    assert type(metrics) is dict, 'provided metrics is not in a dict'
    for metric in metrics:
        try:
            assert isinstance(metric, string_types), "metrics key {} is not a string".format(str(metric))
            experiment_utils._validate_optimization_value(metrics[metric])
        except ValueError:
            raise AssertionError("{} is not a number, only numbers can be attached as metadata for models.".format(str(metrics[metric])))

class Metric:
    MAX = "MAX"
    MIN = "MIN"

class ModelNotFound(Exception):
    """This exception will be raised if the requested model could not be found"""


class ModelArchiveExists(Exception):
    """This exception will be raised if the model archive already exists."""


class MultipleModelFilesFound(Exception):
    """This exception will be raised if multiple model files are found"""



"""
Gridsearch implementation
"""

from hops import hdfs, tensorboard, devices

from hops.experiment_impl.util import experiment_utils
from hops.experiment import Direction

import threading
import six
import time
import os

def _run(sc, train_fn, run_id, args_dict, direction=Direction.MAX, local_logdir=False, name="no-name", optimization_key=None):
    """
    Run the wrapper function with each hyperparameter combination as specified by the dictionary

    Args:
        sc:
        train_fn:
        args_dict:
        direction:
        local_logdir:
        name:

    Returns:

    """
    app_id = str(sc.applicationId)
    num_executions = 1

    if direction.upper() != Direction.MAX and direction.upper() != Direction.MIN:
        raise ValueError('Invalid direction ' + direction +  ', must be Direction.MAX or Direction.MIN')

    arg_lists = list(args_dict.values())
    currentLen = len(arg_lists[0])
    for i in range(len(arg_lists)):
        if currentLen != len(arg_lists[i]):
            raise ValueError('Length of each function argument list must be equal')
        num_executions = len(arg_lists[i])

    #Each TF task should be run on 1 executor
    nodeRDD = sc.parallelize(range(num_executions), num_executions)

    #Make SparkUI intuitive by grouping jobs
    sc.setJobGroup(os.environ['ML_ID'], "{} | Grid Search".format(name))

    #Force execution on executor, since GPU is located on executor
    nodeRDD.foreachPartition(_prepare_func(app_id, run_id, train_fn, args_dict, local_logdir, optimization_key))

    arg_count = six.get_function_code(train_fn).co_argcount
    arg_names = six.get_function_code(train_fn).co_varnames
    exp_dir = experiment_utils._get_logdir(app_id, run_id)

    max_val, max_hp, min_val, min_hp, avg, max_return_dict, min_return_dict = experiment_utils._get_best(args_dict, num_executions, arg_names, arg_count, exp_dir, optimization_key)

    param_combination = ""
    best_val = ""
    return_dict = {}

    if direction.upper() == Direction.MAX:
        param_combination = max_hp
        best_val = str(max_val)
        return_dict = max_return_dict
    elif direction.upper() == Direction.MIN:
        param_combination = min_hp
        best_val = str(min_val)
        return_dict = min_return_dict

    print('Finished Experiment \n')

    best_dir = exp_dir + '/' + param_combination

    return best_dir, experiment_utils._get_params_dict(best_dir), best_val, return_dict

def _prepare_func(app_id, run_id, train_fn, args_dict, local_logdir, optimization_key):
    """

    Args:
        app_id:
        run_id:
        train_fn:
        args_dict:
        local_logdir:

    Returns:

    """

    def _wrapper_fun(iter):
        """

        Args:
            iter:

        Returns:

        """

        for i in iter:
            executor_num = i

        experiment_utils._set_ml_id(app_id, run_id)

        tb_hdfs_path = ''
        hdfs_exec_logdir = ''

        t = threading.Thread(target=devices._print_periodic_gpu_utilization)
        if devices.get_num_gpus() > 0:
            t.start()

        try:
            #Arguments
            if args_dict:
                param_string, params, args = experiment_utils.build_parameters(train_fn, executor_num, args_dict)
                hdfs_exec_logdir, hdfs_appid_logdir = experiment_utils._create_experiment_subdirectories(app_id, run_id, param_string, 'grid_search', params=params)
                logfile = experiment_utils._init_logger(hdfs_exec_logdir)
                tb_hdfs_path, tb_pid = tensorboard._register(hdfs_exec_logdir, hdfs_appid_logdir, executor_num, local_logdir=local_logdir)
                print(devices._get_gpu_info())
                print('-------------------------------------------------------')
                print('Started running task ' + param_string)
                task_start = time.time()
                retval = train_fn(*args)
                task_end = time.time()
                experiment_utils._handle_return(retval, hdfs_exec_logdir, optimization_key, logfile)
                time_str = 'Finished task ' + param_string + ' - took ' + experiment_utils._time_diff(task_start, task_end)
                print(time_str)
                print('Returning metric ' + str(retval))
                print('-------------------------------------------------------')
        except:
            raise
        finally:
            experiment_utils._cleanup(tensorboard, t)

    return _wrapper_fun


import os
import dns.resolver

class ServiceDiscovery:

    @staticmethod
    def get_any_service(service_name):
        return ServiceDiscovery.get_service(service_name)[0]
        
    @staticmethod
    def get_service(service_name):
        service_fqdn = ServiceDiscovery.construct_service_fqdn(service_name)
        answer = dns.resolver.query(service_fqdn, 'SRV')
        return [(a.target.to_text(), a.port) for a in answer]

    @staticmethod
    def construct_service_fqdn(service_name):
        consul_domain = os.getenv('SERVICE_DISCOVERY_DOMAIN', "consul")
        if service_name.endswith('.'):
            return service_name + "service." + consul_domain
        return service_name + ".service." + consul_domain



"""
Utility functions to export models to the Models dataset and get information about models currently being served
in the project.
"""

import os
import json
import re

from hops import hdfs, constants, util, exceptions, kafka

import logging
log = logging.getLogger(__name__)


def exists(serving_name):
    """
    Checks if there exists a serving with the given name

    Example use-case:

    >>> from hops import serving
    >>> serving.exist(serving_name)

    Args:
        :serving_name: the name of the serving

    Returns:
           True if the serving exists, otherwise false
    """
    try:
        return get_id(serving_name) is not None
    except ServingNotFound:
        log.info("No serving with name {} was found in the project {}".format(serving_name, hdfs.project_name()))
        return False


def delete(serving_name):
    """
    Deletes serving instance with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.delete("irisFlowerClassifier")

    Args:
        :serving_name: name of the serving to delete

    Returns:
        None
    """
    serving_id = get_id(serving_name)
    log.debug("Deleting serving with name: {}...".format(serving_name))
    _delete_serving_rest(serving_id)
    log.info("Serving with name: {} successfully deleted".format(serving_name))


def _delete_serving_rest(serving_id):
    """
    Makes a REST request to Hopsworks REST API for deleting a serving instance

    Args:
        :serving_id: id of the serving to delete

    Returns:
        None

    Raises:
        :RestAPIError: if there was an error with the REST call to Hopsworks
    """
    method = constants.HTTP_CONFIG.HTTP_DELETE
    resource_url = (constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_SERVING_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER
                    + str(serving_id))
    response = util.send_request(method, resource_url)

    if response.status_code != 200:
        response_object = response.json()
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise exceptions.RestAPIError("Could not delete serving with id {} (url: {}), "
                                      "server response: \n "
                                      "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, "
                                      "user msg: {}".format(serving_id, resource_url, response.status_code,
                                                            response.reason, error_code, error_msg, user_msg))


def start(serving_name):
    """
    Starts a model serving instance with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.start("irisFlowerClassifier")

    Args:
        :serving_name: name of the serving to start

    Returns:
        None
    """
    serving_id = get_id(serving_name)
    log.debug("Starting serving with name: {}...".format(serving_name))
    _start_or_stop_serving_rest(serving_id, constants.MODEL_SERVING.SERVING_ACTION_START)
    log.info("Serving with name: {} successfully started".format(serving_name))


def stop(serving_name):
    """
    Stops a model serving instance with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.stop("irisFlowerClassifier")

    Args:
        :serving_name: name of the serving to stop

    Returns:
        None
    """
    serving_id = get_id(serving_name)
    log.debug("Stopping serving with name: {}...".format(serving_name))
    _start_or_stop_serving_rest(serving_id, constants.MODEL_SERVING.SERVING_ACTION_STOP)
    log.info("Serving with name: {} successfully stopped".format(serving_name))


def _start_or_stop_serving_rest(serving_id, action):
    """
    Makes a REST request to Hopsworks REST API for starting/stopping a serving instance

    Args:
        :serving_id: id of the serving to start/stop
        :action: the action to perform (start or stop)

    Returns:
        None

    Raises:
        :RestAPIError: if there was an error with the REST call to Hopsworks
    """
    method = constants.HTTP_CONFIG.HTTP_POST
    resource_url = (constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_SERVING_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER
                    + str(serving_id) + constants.MODEL_SERVING.SERVING_START_OR_STOP_PATH_PARAM + action)
    response = util.send_request(method, resource_url)

    if response.status_code != 200:
        response_object = response.json()
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise exceptions.RestAPIError("Could not perform action {} on serving with id {} (url: {}), "
                                      "server response: \n "
                                      "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, "
                                      "user msg: {}".format(action, serving_id, resource_url, response.status_code,
                                                            response.reason, error_code, error_msg, user_msg))


def create_or_update(serving_name, model_path, model_server, model_version=1, artifact_version=None, predictor=None, transformer=None, kfserving=None,
                     batching_enabled=False, topic_name="CREATE", num_partitions=1, num_replicas=1, inference_logging=constants.MODEL_SERVING.INFERENCE_LOGGING_ALL,
                     instances=1, transformer_instances=None, predictor_resource_config=None):
    """
    Creates a serving in Hopsworks if it does not exist, otherwise update the existing one.
    In case model server is not specified, it is inferred from the artifact files.
    If a transformer is specified, KFServing is enabled by default.

    Example use-case:

    >>> from hops import serving
    >>> serving.create_or_update("mnist", "/Models/mnist")

    Args:
        :serving_name: name of the serving to create
        :model_path: path to the model directory
        :model_server: name of the model server to deploy, e.g "TENSORFLOW_SERVING" or "PYTHON"
        :model_version: version of the model to serve
        :artifact_version: version of the artifact to serve (Kubernetes only), e.g "CREATE", "MODEL-ONLY" or version number.
        :predictor: path to the predictor script (python script implementing the Predict class).
        :transformer: path to the transformer script (python script implementing the Transformer class).
        :kfserving: boolean flag whether to serve the model using KFServing serving tool
        :batching_enabled: boolean flag whether to enable batching for the inference requests
        :topic_name: name of the kafka topic for inference logging, e.g "CREATE" to create a new one, "NONE" to not use kafka topic or an existent topic name
        :num_partitions: if a new kafka topic is to created, number of partitions of the new topic
        :num_replicas: if a new kafka topic is to created, replication factor of the new topic
        :inference_logging: inference data to log into the Kafka topic, e.g "MODEL_INPUTS", "PREDICTIONS" or "ALL"
        :instances: the number of serving instances (the more instances the more inference requests can
        be served in parallel)
        :transformer_instances: the number of serving instances (the more instances the more inference requests can
        be served in parallel)
        :predictor_resource_config: dict for setting resource configuration parameters required to serve the model, for
        example {'memory': 2048, 'cores': 1, 'gpus': 0}. Currently only supported if Hopsworks is deployed with Kubernetes installed.

    Returns:
          None
    """
    model_path = hdfs._expand_path(model_path)
    if transformer is not None and kfserving is None:
        kfserving = True

    _validate_user_serving_input(serving_name, model_path, model_server, model_version, artifact_version, predictor, transformer, kfserving,
                                 batching_enabled, topic_name, num_partitions, num_replicas, inference_logging, instances, transformer_instances,
                                 predictor_resource_config)
    model_path = hdfs.get_plain_path(model_path)
    serving_id = get_id(serving_name)
    log.debug("Creating serving {} for artifact {} ...".format(serving_name, model_path))
    _create_or_update_serving_rest(serving_id, serving_name, model_path, model_server, model_version, artifact_version, predictor, transformer, kfserving,
                                   batching_enabled, topic_name, num_partitions, num_replicas, inference_logging, instances, transformer_instances,
                                   predictor_resource_config)
    log.info("Serving {} successfully created".format(serving_name))


def _validate_user_serving_input(serving_name, model_path, model_server, model_version, artifact_version, predictor, transformer, kfserving,
                                 batching_enabled, topic_name, num_partitions, num_replicas, inference_logging, instances, transformer_instances,
                                 predictor_resource_config):
    """
    Validate user input on the client side before sending REST call to Hopsworks (additional validation will be done
    in the backend)

    Args:
        :serving_name: the name of the serving to create
        :model_path: path to the model directory
        :model_server: name of the model server to deploy, e.g "TENSORFLOW_SERVING" or "PYTHON"
        :model_version: version of the model to serve
        :artifact_version: version of the artifact to serve
        :predictor: path to the predictor script
        :transformer: path to the transformer script
        :kfserving: boolean flag whether to serve the model using KFServing serving tool
        :batching_enabled: boolean flag whether to enable batching for inference requests to the serving
        :topic_name: name of the kafka topic for inference logging, e.g "CREATE" to create a new one, "NONE" to not use kafka topic or an existent topic name
        :num_partitions: if a new kafka topic is to created, number of partitions of the new topic
        :num_replicas: if a new kafka topic is to created, replication factor of the new topic
        :inference_logging: inference data to log into the Kafka topic, e.g "MODEL_INPUTS", "PREDICTIONS" or "ALL"
        :instances: the number of serving instances (the more instances the more inference requests can
        be served in parallel)
        :transformer_instances: the number of transformer instances (the more instances the more inference requests can
        be served in parallel)
        :predictor_resource_config: dict for setting resource configuration parameters required to serve the model, for
        example {'memory': 2048, 'cores': 1, 'gpus': 0}. Currently only supported if Hopsworks is deployed with Kubernetes installed.

    Returns:
        None

    Raises:
        :ValueError: if the serving input failed the validation
    """
    name_pattern = re.compile("^[a-zA-Z0-9]+$")
    if len(serving_name) > 256 or serving_name == "" or not name_pattern.match(serving_name):
        raise ValueError("Name of serving cannot be empty, cannot exceed 256 characters and must match the regular "
                         "expression: ^[a-zA-Z0-9]+$, the provided name: {} is not valid".format(serving_name))
    if not hdfs.exists(model_path):
        raise ValueError("The model/artifact path must exist in HDFS, the provided path: {} "
                         "does not exist".format(model_path))
    if model_server is None:
        raise ValueError("Model server not provided, supported "
                         "model servers are: {}".format(",".join(constants.MODEL_SERVING.MODEL_SERVERS)))
    if model_server not in constants.MODEL_SERVING.MODEL_SERVERS:
        raise ValueError("The provided model_server: {} is not supported, supported "
                         "model servers are: {}".format(model_server, ",".join(constants.MODEL_SERVING.MODEL_SERVERS)))
    if inference_logging is not None and inference_logging not in constants.MODEL_SERVING.INFERENCE_LOGGING_MODES:
        raise ValueError("The provided inference_logging: {} is not supported, supported "
                         "inference logging modes are: {}".format(inference_logging, ",".join(constants.MODEL_SERVING.INFERENCE_LOGGING_MODES)))
    if not isinstance(model_version, int):
        raise ValueError("The model version must be an integer, the provided version is not: {}".format(model_version))
    if model_server == constants.MODEL_SERVING.MODEL_SERVER_TENSORFLOW_SERVING:
        if not isinstance(num_replicas, int):
            raise ValueError("Number of kafka topic replicas must be an integer, the provided num replicas "
                             "is not: {}".format(model_version))
        if not isinstance(num_partitions, int):
            raise ValueError("Number of kafka topic partitions must be an integer, the provided num partitions "
                             "is not: {}".format(num_partitions))
        if not isinstance(batching_enabled, bool):
            raise ValueError("Batching enabled must be a boolean, the provided value "
                             "is not: {}".format(batching_enabled))
        if predictor is not None:
            raise ValueError("Predictors are not supported with Tensorflow Serving")
        if kfserving and batching_enabled:
            raise ValueError("Batching requests is currently not supported in KFServing deployments")

    if not isinstance(instances, int):
        raise ValueError("The number of serving instances must be an integer, "
                         "the provided version is not: {}".format(instances))

    if not kfserving:
        if inference_logging is not None and inference_logging != constants.MODEL_SERVING.INFERENCE_LOGGING_ALL:
            raise ValueError("Fine-grained inference logging is only supported in KFServing deployments")
        if topic_name is not None and topic_name != "NONE" and inference_logging != constants.MODEL_SERVING.INFERENCE_LOGGING_ALL:
            raise ValueError("Inference logging mode 'ALL' is the only mode supported for non-KFServing deployments")

    if kfserving:
        if topic_name is not None and topic_name != "NONE" and inference_logging is None:
            raise ValueError("Inference logging must be defined. Supported inference "
                             "logging modes are: {}".format(",".join(constants.MODEL_SERVING.INFERENCE_LOGGING_MODES)))

    if predictor_resource_config is not None:
        if type(predictor_resource_config) is not dict:
            raise ValueError("predictor_resource_config must be a dict.")
        if 'memory' not in predictor_resource_config or 'cores' not in predictor_resource_config:
            raise ValueError("predictor_resource_config must contain the keys 'memory' and 'cores'")


def _create_or_update_serving_rest(serving_id, serving_name, model_path, model_server, model_version, artifact_version, predictor, transformer, kfserving,
                                   batching_enabled, topic_name, num_partitions, num_replicas, inference_logging, instances, transformer_instances,
                                   predictor_resource_config):
    """
    Makes a REST request to Hopsworks for creating or updating a model serving instance

    Args:
        :serving_id: the id of the serving in case of UPDATE, if serving_id is None, it is a CREATE operation.
        :serving_name: the name of the serving to create
        :model_path: path to the model directory
        :model_server: name of the model server to deploy, e.g "TENSORFLOW_SERVING" or "PYTHON"
        :model_version: version of the model to serve
        :artifact_version: version of the artifact to serve
        :predictor: path to the predictor script
        :transformer: path to the transformer script
        :kfserving: boolean flag whether to serve the model using KFServing serving tool
        :batching_enabled: boolean flag whether to enable batching for inference requests to the serving
        :topic_name: name of the kafka topic for inference logging, e.g "CREATE" to create a new one, "NONE" to not use kafka topic or an existent topic name
        :num_partitions: if a new kafka topic is to created, number of partitions of the new topic
        :num_replicas: if a new kafka topic is to created, replication factor of the new topic
        :inference_logging: inference data to log into the Kafka topic, e.g "MODEL_INPUTS", "PREDICTIONS" or "ALL"
        :instances: the number of serving instances (the more instances the more inference requests can
        be served in parallel)
        :transformer_instances: the number of transformer instances (the more instances the more inference requests can
        be served in parallel)
        :predictor_resource_config: dict for setting resource configuration parameters required to serve the model, for
        example {'memory': 2048, 'cores': 1, 'gpus': 0}. Currently only supported if Hopsworks is deployed with Kubernetes installed.

    Returns:
        None

    Raises:
        :RestAPIError: if there was an error with the REST call to Hopsworks
    """

    serving_tool = constants.MODEL_SERVING.SERVING_TOOL_KFSERVING if kfserving else constants.MODEL_SERVING.SERVING_TOOL_DEFAULT

    json_contents = {
        constants.REST_CONFIG.JSON_SERVING_NAME: serving_name,
        constants.REST_CONFIG.JSON_SERVING_MODEL_PATH: model_path,
        constants.REST_CONFIG.JSON_SERVING_MODEL_VERSION: model_version,
        constants.REST_CONFIG.JSON_SERVING_ARTIFACT_VERSION: artifact_version,
        constants.REST_CONFIG.JSON_SERVING_PREDICTOR: predictor,
        constants.REST_CONFIG.JSON_SERVING_TRANSFORMER: transformer,
        constants.REST_CONFIG.JSON_SERVING_MODEL_SERVER: model_server,
        constants.REST_CONFIG.JSON_SERVING_TOOL: serving_tool,
        constants.REST_CONFIG.JSON_SERVING_KAFKA_TOPIC_DTO: {
            constants.REST_CONFIG.JSON_KAFKA_TOPIC_NAME: topic_name,
            constants.REST_CONFIG.JSON_KAFKA_NUM_PARTITIONS: num_partitions,
            constants.REST_CONFIG.JSON_KAFKA_NUM_REPLICAS: num_replicas
        },
        constants.REST_CONFIG.JSON_SERVING_INFERENCE_LOGGING: inference_logging,
        constants.REST_CONFIG.JSON_SERVING_REQUESTED_INSTANCES: instances,
        constants.REST_CONFIG.JSON_SERVING_REQUESTED_TRANSFORMER_INSTANCES: transformer_instances,
        constants.REST_CONFIG.JSON_SERVING_PREDICTOR_RESOURCE_CONFIG: predictor_resource_config
    }
    if serving_id is not None:
        json_contents[constants.REST_CONFIG.JSON_SERVING_ID] = serving_id
    if model_server == constants.MODEL_SERVING.MODEL_SERVER_TENSORFLOW_SERVING:
        json_contents[constants.REST_CONFIG.JSON_SERVING_BATCHING_ENABLED] = batching_enabled
    if artifact_version == "CREATE":
        json_contents[constants.REST_CONFIG.JSON_SERVING_ARTIFACT_VERSION] = -1
    elif artifact_version == "MODEL-ONLY":
        json_contents[constants.REST_CONFIG.JSON_SERVING_ARTIFACT_VERSION] = 0
    if topic_name is None or topic_name == "NONE":
        json_contents[constants.REST_CONFIG.JSON_SERVING_INFERENCE_LOGGING] = None

    json_embeddable = json.dumps(json_contents)
    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    method = constants.HTTP_CONFIG.HTTP_PUT
    resource_url = (constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_SERVING_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER)
    response = util.send_request(method, resource_url, data=json_embeddable, headers=headers)

    if response.status_code != 201 and response.status_code != 200:
        response_object = response.json()
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise exceptions.RestAPIError("Could not create or update serving (url: {}), server response: \n "
                                      "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, "
                                      "user msg: {}".format(resource_url, response.status_code, response.reason,
                                                            error_code, error_msg, user_msg))


def get(serving_name):
    """
    Gets the serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get(serving_name)

    Args:
        :serving_name: name of the serving to get

    Returns:
         the serving, None if Serving does not exist
    """
    try:
        servings = get_all()
        serving = _find_serving_with_name(serving_name, servings)
        return serving
    except ServingNotFound:
        return None


def get_id(serving_name):
    """
    Gets the id of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_id(serving_name)

    Args:
        :serving_name: name of the serving to get the id for

    Returns:
         the id of the serving, None if Serving does not exist
    """
    try:
        servings = get_all()
        serving = _find_serving_with_name(serving_name, servings)
        return serving.id
    except ServingNotFound:
        return None


def get_model_path(serving_name):
    """
    Gets the artifact path of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_model_path(serving_name)

    Args:
        :serving_name: name of the serving to get the artifact path for

    Returns:
         the artifact path of the serving (model path in case of tensorflow, or python script in case of SkLearn)
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.model_path


def get_model_name(serving_name):
    """
    Gets the name of the model served by a given serving instance

    Example use-case:

    >>> from hops import serving
    >>> serving.get_model_name(serving_name)

    Args:
        :serving_name: name of the serving to get the model name for

    Returns:
         the name of the model being served
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.model_version


def get_model_version(serving_name):
    """
    Gets the version of the model served by a given serving instance

    Example use-case:

    >>> from hops import serving
    >>> serving.get_model_version(serving_name)

    Args:
        :serving_name: name of the serving to get the version for

    Returns:
         the version of the model being served
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.model_version


def get_artifact_version(serving_name):
    """
    Gets the version of the artifact served by a given serving instance

    Example use-case:

    >>> from hops import serving
    >>> serving.get_artifact_version(serving_name)

    Args:
        :serving_name: name of the serving to get the version for

    Returns:
         the version of the artifact being served
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.artifact_version


def get_predictor(serving_name):
    """
    Gets the predictor filename used in a given serving instance

    Example use-case:

    >>> from hops import serving
    >>> serving.get_predictor(serving_name)

    Args:
        :serving_name: name of the serving to get the predictor for

    Returns:
         the predictor filename
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.predictor


def get_transformer(serving_name):
    """
    Gets the transformer filename used in a given serving instance

    Example use-case:

    >>> from hops import serving
    >>> serving.get_transformer(serving_name)

    Args:
        :serving_name: name of the serving to get the transformer for

    Returns:
         the transformer filename
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.transformer


def get_model_server(serving_name):
    """
    Gets the type of model server of the serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_model_server(serving_name)

    Args:
        :serving_name: name of the serving

    Returns:
         the model server (e.g Tensorflow Serving or Python)
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.model_server


def get_serving_tool(serving_name):
    """
    Gets the serving tool of the serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_serving_tool(serving_name)

    Args:
        :serving_name: name of the serving

    Returns:
         the serving tool (e.g DEFAULT or KFSERVING)
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.serving_tool


def get_available_instances(serving_name):
    """
    Gets the number of available instances of the serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_available_instances(serving_name)

    Args:
        :serving_name: name of the serving

    Returns:
         number of available replicas (e.g int or (int, int) if the serving includes a transformer)
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.available_instances if serving.available_transformer_instances is None \
        else (serving.available_instances, serving.available_transformer_instances)


def get_kafka_topic(serving_name):
    """
    Gets the kafka topic name of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_kafka_topic(serving_name)

    Args:
        :serving_name: name of the serving to get the kafka topic name for

    Returns:
         the kafka topic name of the serving
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.kafka_topic_dto.name


def get_status(serving_name):
    """
    Gets the status of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_status(serving_name)

    Args:
        :serving_name: name of the serving to get the status for

    Returns:
         the status of the serving
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.status


def get_predictor_resource_config(serving_name):
    """
    Gets the predictor resource config of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_predictor_resource_config(serving_name)

    Args:
        :serving_name: name of the serving to get the predictor resource config for

    Returns:
         the status of the serving
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    return serving.predictor_resource_config


def get_internal_endpoints(serving_name):
    """
    Gets the internal endpoints of a serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.get_internal_endpoints(serving_name)

    Args:
        :serving_name: name of the serving to get the internal endoints for

    Returns:
         internal endpoints
    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)    
    if serving.internal_path is None:
        return None

    protocol = "http"
    endpoints = []
    for ip in serving.internal_ips:
        endpoints.append(protocol + "://" + ip + ":" + serving.internal_port)

    return endpoints


def describe(serving_name):
    """
    Describes the serving with a given name

    Example use-case:

    >>> from hops import serving
    >>> serving.describe(serving_name)

    Args:
        :serving_name: name of the serving

    """
    servings = get_all()
    serving = _find_serving_with_name(serving_name, servings)
    log.info(', '.join("%s: %s" % item for item in vars(serving).items()))


def get_all():
    """
    Gets the list of servings for the current project

    Example:

    >>> from hops import serving
    >>> servings = serving.get_all()
    >>> servings[0].name

    Returns:
         list of servings
    """
    return _parse_json_servings(_get_servings_rest())


def _find_serving_with_name(serving_name, servings):
    """
    Finds a serving with a given name from a list of servings (O(N))

    Args:
        :serving_name: name of the serving to look for
        :servings: the list of servings to look through

    Returns:
           serving with the given name

    Raises:
        :ServingNotFound: if the requested serving could not be found
    """
    serving_names = []
    for serving in servings:
        if serving.name == serving_name:
            return serving
        serving_names.append(serving.name)
    serving_names_str = ",".join(serving_names)
    raise ServingNotFound("No serving with name: {} could be found among the list of "
                          "available servings: {}".format(serving_name, serving_names_str))


def _parse_json_servings(json_servings):
    """
    Parses a list of JSON servings into Serving Objects

    Args:
        :json_servings: the list of JSON servings

    Returns:
           a list of Serving Objects
    """
    return list(map(lambda json_serving: Serving(json_serving), json_servings))


def _get_servings_rest():
    """
    Makes a REST request to Hopsworks to get a list of all servings in the current project

    Returns:
         JSON response parsed as a python dict

    Raises:
        :RestAPIError: if there was an error with the REST call to Hopsworks
    """
    method = constants.HTTP_CONFIG.HTTP_GET
    resource_url = (constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_SERVING_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER)
    response = util.send_request(method, resource_url)
    response_object = response.json()
    if response.status_code != 200:
        error_code, error_msg, user_msg = util._parse_rest_error(response_object)
        raise exceptions.RestAPIError("Could not fetch list of servings from Hopsworks REST API (url: {}), "
                                      "server response: \n "
                                      "HTTP code: {}, HTTP reason: {}, error code: {}, "
                                      "error msg: {}, user msg: {}".format(resource_url, response.status_code, response.reason,
                                                                           error_code, error_msg, user_msg))
    return response_object


def make_inference_request(serving_name, data, verb=":predict"):
    """
    Submit an inference request

    Example use-case:

    >>> from hops import serving
    >>> serving.make_inference_request("irisFlowerClassifier", [[1,2,3,4]], ":predict")

    Args:
        :serving_name: name of the model being served
        :data: data/json to send to the serving
        :verb: type of request (:predict, :classify, or :regress)

    Returns:
        the JSON response
    """
    return _make_inference_request_rest(serving_name, data, verb)


def _make_inference_request_rest(serving_name, data, verb):
    """
    Makes a REST request to Hopsworks for submitting an inference request to the serving instance

    Args:
        :serving_name: name of the model being served
        :data: data/json to send to the serving
        :verb: type of request (:predict, :classify, or :regress)

    Returns:
        the JSON response

    Raises:
        :RestAPIError: if there was an error with the REST call to Hopsworks
    """
    json_embeddable = json.dumps(data)

    headers = {constants.HTTP_CONFIG.HTTP_CONTENT_TYPE: constants.HTTP_CONFIG.HTTP_APPLICATION_JSON}
    if constants.ENV_VARIABLES.SERVING_API_KEY_ENV_VAR in os.environ:
        headers[constants.HTTP_CONFIG.HTTP_AUTHORIZATION] = "ApiKey " + os.environ[constants.ENV_VARIABLES.SERVING_API_KEY_ENV_VAR]

    method = constants.HTTP_CONFIG.HTTP_POST
    resource_url = (constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_PROJECT_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    hdfs.project_id() + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_INFERENCE_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    constants.REST_CONFIG.HOPSWORKS_MODELS_RESOURCE + constants.DELIMITERS.SLASH_DELIMITER +
                    serving_name + verb)
    response = util.send_request(method, resource_url, data=json_embeddable, headers=headers)
    response_object = response.json()
    error_code, error_msg, user_msg = util._parse_rest_error(response_object)

    if response.status_code != 201 and response.status_code != 200:
        raise exceptions.RestAPIError("Could not create or update serving (url: {}), server response: \n "
                                      "HTTP code: {}, HTTP reason: {}, error code: {}, error msg: {}, "
                                      "user msg: {}".format(resource_url, response.status_code, response.reason,
                                                            error_code, error_msg, user_msg))
    return response_object


class Serving(object):
    """
    Represents a model being served in Hopsworks
    """

    def __init__(self, serving_json):
        """
        Initialize the serving from JSON payload returned by Hopsworks REST API

        Args:
            :feature_json: JSON data about the feature returned from Hopsworks REST API
        """
        self.id = serving_json[constants.REST_CONFIG.JSON_SERVING_ID]
        self.name = serving_json[constants.REST_CONFIG.JSON_SERVING_NAME]
        self.model_path = serving_json[constants.REST_CONFIG.JSON_SERVING_MODEL_PATH]
        self.model_name = serving_json[constants.REST_CONFIG.JSON_SERVING_MODEL_NAME]
        self.model_version = serving_json[constants.REST_CONFIG.JSON_SERVING_MODEL_VERSION]
        self.artifact_version = serving_json[constants.REST_CONFIG.JSON_SERVING_ARTIFACT_VERSION] \
            if constants.REST_CONFIG.JSON_SERVING_ARTIFACT_VERSION in serving_json \
            else None
        self.predictor = serving_json[constants.REST_CONFIG.JSON_SERVING_PREDICTOR] \
            if constants.REST_CONFIG.JSON_SERVING_PREDICTOR in serving_json \
            else None
        self.transformer = serving_json[constants.REST_CONFIG.JSON_SERVING_TRANSFORMER] \
            if constants.REST_CONFIG.JSON_SERVING_TRANSFORMER in serving_json \
            else None
        self.model_server = serving_json[constants.REST_CONFIG.JSON_SERVING_MODEL_SERVER]
        self.serving_tool = serving_json[constants.REST_CONFIG.JSON_SERVING_TOOL]
        self.requested_instances = serving_json[constants.REST_CONFIG.JSON_SERVING_REQUESTED_INSTANCES]
        self.available_instances = serving_json[constants.REST_CONFIG.JSON_SERVING_AVAILABLE_INSTANCES] \
            if constants.REST_CONFIG.JSON_SERVING_AVAILABLE_INSTANCES in serving_json \
            else 0
        self.available_transformer_instances = serving_json[constants.REST_CONFIG.JSON_SERVING_AVAILABLE_TRANSFORMER_INSTANCES] \
            if constants.REST_CONFIG.JSON_SERVING_AVAILABLE_TRANSFORMER_INSTANCES in serving_json \
            else None
        if constants.REST_CONFIG.JSON_SERVING_PREDICTOR_RESOURCE_CONFIG in serving_json:
            self.predictor_resource_config = serving_json[constants.REST_CONFIG.JSON_SERVING_PREDICTOR_RESOURCE_CONFIG]
        self.creator = serving_json[constants.REST_CONFIG.JSON_SERVING_CREATOR]
        self.created = serving_json[constants.REST_CONFIG.JSON_SERVING_CREATED]
        self.status = serving_json[constants.REST_CONFIG.JSON_SERVING_STATUS]

        if constants.REST_CONFIG.JSON_SERVING_KAFKA_TOPIC_DTO in serving_json:
            self.kafka_topic_dto = kafka.KafkaTopicDTO(serving_json[constants.REST_CONFIG.JSON_SERVING_KAFKA_TOPIC_DTO])

        if constants.REST_CONFIG.JSON_SERVING_EXTERNAL_IP in serving_json:
            self.external_ip = serving_json[constants.REST_CONFIG.JSON_SERVING_EXTERNAL_IP]
        if constants.REST_CONFIG.JSON_SERVING_EXTERNAL_PORT in serving_json:
            self.external_port = serving_json[constants.REST_CONFIG.JSON_SERVING_EXTERNAL_PORT]
        if constants.REST_CONFIG.JSON_SERVING_INTERNAL_IPS in serving_json:
            self.internal_ips = serving_json[constants.REST_CONFIG.JSON_SERVING_INTERNAL_IPS]
        if constants.REST_CONFIG.JSON_SERVING_INTERNAL_PORT in serving_json:
            self.internal_port = serving_json[constants.REST_CONFIG.JSON_SERVING_INTERNAL_PORT]
        if constants.REST_CONFIG.JSON_SERVING_INTERNAL_PATH in serving_json:
            self.internal_path = serving_json[constants.REST_CONFIG.JSON_SERVING_INTERNAL_PATH]


class ServingNotFound(Exception):
    """This exception will be raised if the requested serving could not be found"""



"""
Differential evolution implementation
"""


import random
from collections import OrderedDict
import os

from hops import hdfs, tensorboard, devices, util
from hops.experiment_impl.util import experiment_utils
from hops.experiment import Direction

import threading
import six
import time
import copy
import json
import sys

objective_function=None
spark=None
opt_key=None
diff_evo=None
cleanup=None
fs_handle=None
local_logdir_bool=False

generation_id = 0
run_id = 0

def _get_all_accuracies(tensorboard_hdfs_logdir, args_dict, number_params):
    """
    Retrieves all accuracies from the parallel executions (each one is in a
    different file, one per combination of wrapper function parameter)

    Args:
        :tensorboard_hdfs_logdir:
        :args_dict:
        :number_params:

    Returns:

    """

    results=[]

    #Important, this must be ordered equally than _parse_to_dict function
    population_dict = diff_evo.get_dict()
    global run_id
    for i in range(number_params):
        path_to_log= tensorboard_hdfs_logdir + "/generation." + str(generation_id - 1) + "/"
        for k in population_dict:
            path_to_log+=k+'='+str(args_dict[k][i])+ '&'
        path_to_log = path_to_log[:(len(path_to_log) -1)]
        path_to_log = path_to_log + '/.metric'

        with hdfs.open_file(path_to_log, flags="r") as fi:
            metric = fi.read()
            fi.close()

        results.append(metric)

    return [float(res) for res in results]

def _execute_all(population_dict, name="no-name"):
    """
    Executes wrapper function with all values of population_dict parallely.
    Returns a list of accuracies (or metric returned in the wrapper) in the
    same order as in the population_dict.

    Args:
        :population_dict:
        :name:

    Returns:

    """
    initial_pop = copy.deepcopy(population_dict)

    number_hp_combinations=[len(v) for v in population_dict.values()][0]
    # directory for current generation


    # Do not run hyperparameter combinations that are duplicated
    # Find all duplicates and delete them
    keys = population_dict.keys()
    i=0
    while i < number_hp_combinations:
        duplicate_entries = _duplicate_entry(i, keys, population_dict, number_hp_combinations)
        if len(duplicate_entries) > 0:
            # sort entries, delete descending
            for index in duplicate_entries:
                for key in keys:
                    population_dict[key].pop(index)
            i=0
            number_hp_combinations = [len(v) for v in population_dict.values()][0]
        else:
            i+=1

    tensorboard_hdfs_logdir = _evolutionary_launch(spark, objective_function, population_dict, name=name)

    return _get_all_accuracies(tensorboard_hdfs_logdir, initial_pop, [len(v) for v in initial_pop.values()][0])


def _duplicate_entry(i, keys, population, len):
    """

    Args:
        :i:
        :keys:
        :population:
        :len:

    Returns:

    """
    hp_combinations = []
    duplicate_indices = []

    for val in range(len):
        entry=''
        for key in keys:
            entry += str(population[key][val]) + '='
        hp_combinations.append(entry)

    to_find = hp_combinations[i]
    #get the duplicate indices
    for y in range(len):
        if hp_combinations[y] == to_find and y != i:
            duplicate_indices.insert(0, y)

    return duplicate_indices

class DifferentialEvolution:
    _types = ['float', 'int', 'cat']
    _generation = 0
    _scores = []
    _ordered_population_dict = []
    _param_names = []

    def __init__(self, objective_function, parbounds, types, ordered_dict, direction=Direction.MAX, generations=10, population=10, mutation=0.5, crossover=0.7, name="no-name"):
        """

        Args:
            :objective_function:
            :parbounds:
            :types:
            :ordered_dict:
            :direction:
            :generations:
            :population:
            :mutation:
            :crossover:
            :name:
        """
        self.objective_function = objective_function
        self.parbounds = parbounds
        self.direction = direction
        self.types = types
        self.generations = generations
        self.n = population
        self.F = mutation
        self.CR = crossover
        self._ordered_population_dict = ordered_dict
        self.name = name

        global generation_id
        generation_id = 0

        self._param_names = []
        for entry in ordered_dict:
            self._param_names.append(entry)

        #self.m = -1 if maximize else 1

    # run differential evolution algorithms
    def _solve(self, root_dir):
        """

        Args:
            :root_dir:

        Returns:

        """
        # initialise generation based on individual representation
        population, bounds = self._population_initialisation()
        global fs_handle
        fs_handle = hdfs.get_fs()
        global run_id

        new_gen_best_param = None
        new_gen_best = None

        for _ in range(self.generations):

            donor_population = self._mutation(population, bounds)
            trial_population = self._recombination(population, donor_population)

            population = self._selection(population, trial_population)

            new_gen_avg = sum(self._scores)/self.n

            if self.direction.upper() == Direction.MAX:
                new_gen_best = max(self._scores)
            elif self.direction.upper() == Direction.MIN:
                new_gen_best = min(self._scores)
            else:
                raise ValueError('invalid direction: ' + self.direction)

            new_gen_best_param = self._parse_back(population[self._scores.index(new_gen_best)])

            index = 0
            for name in self._param_names:
                new_gen_best_param[index] = name + "=" + str(new_gen_best_param[index])
                index += 1

            print("Generation " + str(self._generation) + " || " + "average metric: " + str(new_gen_avg) \
                  + ", best metric: " + str(new_gen_best) + ", best parameter combination: " + str(new_gen_best_param) + "\n")

            if cleanup:
                hdfs.rmr(root_dir + '/generation.' + str(self._generation-1))

        parsed_back_population = []
        for indiv in population:
            parsed_back_population.append(self._parse_back(indiv))

        return new_gen_best_param, new_gen_best

    # define bounds of each individual depending on type
    def _individual_representation(self):
        """

        Returns:

        """
        bounds = []

        for index, item in enumerate(self.types):
            b =()
            # if categorical then take bounds from 0 to number of items
            if item == self._types[2]:
                b = (0, int(len(self.parbounds[index]) - 1))
            # if float/int then take given bounds
            else:
                b = self.parbounds[index]
            bounds.append(b)
        return bounds

    # initialise population
    def _population_initialisation(self):
        """

        Returns:

        """
        population = []
        num_parameters = len(self.parbounds)
        for i in range(self.n):
            indiv = []
            bounds = self._individual_representation()

            for i in range(num_parameters):
                indiv.append(random.uniform(bounds[i][0], bounds[i][1]))
            indiv = self._ensure_bounds(indiv, bounds)
            population.append(indiv)
        return population, bounds

    # ensure that any mutated individual is within bounds
    def _ensure_bounds(self, indiv, bounds):
        """

        Args:
            :indiv:
            :bounds:

        Returns:

        """
        indiv_correct = []

        for i in range(len(indiv)):
            par = indiv[i]

            # check if param is within bounds
            lowerbound = bounds[i][0]
            upperbound = bounds[i][1]
            if par < lowerbound:
                par = lowerbound
            elif par > upperbound:
                par = upperbound

            # check if param needs rounding
            if self.types[i] != 'float':
                par = int(round(par))
            indiv_correct.append(par)
        return indiv_correct

    # create donor population based on mutation of three vectors
    def _mutation(self, population, bounds):
        """

        Args:
            :population:
            :bounds:

        Returns:

        """
        donor_population = []
        for i in range(self.n):

            indiv_indices = list(range(0, self.n))
            indiv_indices.remove(i)

            candidates = random.sample(indiv_indices, 3)
            x_1 = population[candidates[0]]
            x_2 = population[candidates[1]]
            x_3 = population[candidates[2]]

            # substracting the second from the third candidate
            x_diff = [x_2_i - x_3_i for x_2_i, x_3_i in zip(x_2, x_3)]
            donor_vec = [x_1_i + self.F*x_diff_i for x_1_i, x_diff_i in zip (x_1, x_diff)]
            donor_vec = self._ensure_bounds(donor_vec, bounds)
            donor_population.append(donor_vec)

        return donor_population

    # recombine donor vectors according to crossover probability
    def _recombination(self, population, donor_population):
        """

        Args:
            :population:
            :donor_population:

        Returns:

        """
        trial_population = []
        for k in range(self.n):
            target_vec = population[k]
            donor_vec = donor_population[k]
            trial_vec = []
            for p in range(len(self.parbounds)):
                crossover = random.random()

                # if random number is below set crossover probability do recombination
                if crossover <= self.CR:
                    trial_vec.append(donor_vec[p])
                else:
                    trial_vec.append(target_vec[p])
            trial_population.append(trial_vec)
        return trial_population

    # select the best individuals from each generation
    def _selection(self, population, trial_population):
        """

        Args:
            :population:
            :trial_population:

        Returns:

        """
        # Calculate trial vectors and target vectors and select next generation

        if self._generation == 0:
            parsed_population = []
            for target_vec in population:
                parsed_target_vec = self._parse_back(target_vec)
                parsed_population.append(parsed_target_vec)

            parsed_population = self._parse_to_dict(parsed_population)
            self._scores = self.objective_function(parsed_population, name=self.name)

            if self.direction.upper() == Direction.MAX:
                new_gen_best = max(self._scores)
            elif self.direction.upper() == Direction.MIN:
                new_gen_best = min(self._scores)
            else:
                raise ValueError('invalid direction: ' + self.direction)

            new_gen_best_param = self._parse_back(population[self._scores.index(new_gen_best)])

            index = 0
            for name in self._param_names:
                new_gen_best_param[index] = name + "=" + str(new_gen_best_param[index])
                index += 1

        parsed_trial_population = []
        for index, trial_vec in enumerate(trial_population):
            parsed_trial_vec = self._parse_back(trial_vec)
            parsed_trial_population.append(parsed_trial_vec)

        parsed_trial_population =  self._parse_to_dict(parsed_trial_population)
        trial_population_scores = self.objective_function(parsed_trial_population, name=self.name)

        for i in range(self.n):
            trial_vec_score_i = trial_population_scores[i]
            target_vec_score_i = self._scores[i]
            if self.direction.upper() == Direction.MAX:
                if trial_vec_score_i > target_vec_score_i:
                    self._scores[i] = trial_vec_score_i
                    population[i] = trial_population[i]
            elif self.direction.upper() == Direction.MIN:
                if trial_vec_score_i < target_vec_score_i:
                    self._scores[i] = trial_vec_score_i
                    population[i] = trial_population[i]

        self._generation += 1

        return population
    # parse the converted values back to original
    def _parse_back(self, individual):
        """

        Args:
            :individual:

        Returns:

        """
        original_representation = []
        for index, parameter in enumerate(individual):
            if self.types[index] == self._types[2]:
                original_representation.append(self.parbounds[index][parameter])
            else:

                original_representation.append(parameter)

        return original_representation

    # for parallelization purposes one can parse the population from a list to a  dictionary format
    # User only has to add the parameters he wants to optimize to population_dict
    def _parse_to_dict(self, population):
        """

        Args:
            :population:

        Returns:

        """

        # reset entries
        for entry in self._ordered_population_dict:
            self._ordered_population_dict[entry] = []


        for indiv in population:
            index = 0
            for param in self._param_names:
                self._ordered_population_dict[param].append(indiv[index])
                index = index + 1

        return self._ordered_population_dict

    def get_dict(self):
        return self._ordered_population_dict

def _run(train_fn, search_dict, direction = Direction.MAX, generations=4, population=6, mutation=0.5, crossover=0.7, cleanup_generations=False, local_logdir=False, name="no-name", optimization_key=None):
    """

    Args:
        :train_fn:
        :search_dict:
        :direction:
        :generations:
        :population:
        :mutation:
        :crossover:
        :cleanup_generations:
        :local_logdir:
        :name:
        :optimization_key:

    Returns:

    """

    global run_id
    global local_logdir_bool
    local_logdir_bool = local_logdir

    global spark
    spark = util._find_spark()

    global objective_function
    objective_function = train_fn

    global cleanup
    cleanup = cleanup_generations

    global opt_key
    opt_key = optimization_key

    argcount = six.get_function_code(train_fn).co_argcount
    arg_names = six.get_function_code(train_fn).co_varnames

    ordered_arr = []

    app_id = spark.sparkContext.applicationId

    arg_lists = list(search_dict.values())
    for i in range(len(arg_lists)):
        if len(arg_lists[i]) != 2:
            raise ValueError('Boundary list must contain exactly two elements, [lower_bound, upper_bound] for float/int '
                             'or [category1, category2] in the case of strings')

    assert population > 3, 'population should be greater than 3'
    assert generations > 1, 'generations should be greater than 1'

    argIndex = 0
    while argcount != 0:
        ordered_arr.append((arg_names[argIndex], search_dict[arg_names[argIndex]]))
        argcount = argcount - 1
        argIndex = argIndex + 1

    ordered_dict = OrderedDict(ordered_arr)

    bounds_list = []
    types_list = []

    for entry in ordered_dict:
        bounds_list.append((ordered_dict[entry][0], ordered_dict[entry][1]))

        if isinstance(ordered_dict[entry][0], int):
            types_list.append('int')
        elif isinstance(ordered_dict[entry][0], float):
            types_list.append('float')
        else:
            types_list.append('cat')

    global diff_evo
    diff_evo = DifferentialEvolution(_execute_all,
                                     bounds_list,
                                     types_list,
                                     ordered_dict,
                                     direction=direction,
                                     generations=generations,
                                     population=population,
                                     crossover=crossover,
                                     mutation=mutation,
                                     name=name)

    root_dir = experiment_utils._get_experiments_dir() + "/" + str(app_id) + "_" + str(run_id)

    best_param, best_metric = diff_evo._solve(root_dir)


    param_string = ''
    for hp in best_param:
        param_string = param_string + hp + '&'
    param_string = param_string[:-1]

    best_exp_logdir, return_dict = _get_best(str(root_dir), direction)

    print('Finished Experiment \n')

    return best_exp_logdir, experiment_utils._get_params_dict(best_exp_logdir), best_metric, return_dict

def _evolutionary_launch(spark, train_fn, args_dict, name="no-name"):
    """ Run the wrapper function with each hyperparameter combination as specified by the dictionary

    Args:
        :spark_session: SparkSession object
        :train_fn: The TensorFlow function to run
        :args_dict: (optional) A dictionary containing hyperparameter values to insert as arguments for each TensorFlow job
    """

    global run_id

    sc = spark.sparkContext
    app_id = str(sc.applicationId)


    arg_lists = list(args_dict.values())
    num_executions = len(arg_lists[0])

    #Each TF task should be run on 1 executor
    nodeRDD = sc.parallelize(range(num_executions), num_executions)

    #Force execution on executor, since GPU is located on executor
    global generation_id
    global run_id

    #Make SparkUI intuitive by grouping jobs
    sc.setJobGroup(os.environ['ML_ID'], "{} | Differential Evolution, Generation: {}".format(name, generation_id))
    nodeRDD.foreachPartition(_prepare_func(app_id, generation_id, train_fn, args_dict, run_id, opt_key))

    generation_id += 1

    return experiment_utils._get_experiments_dir() + '/' + str(app_id) + "_" + str(run_id)


#Helper to put Spark required parameter iter in function signature
def _prepare_func(app_id, generation_id, train_fn, args_dict, run_id, opt_key):
    """

    Args:
        :app_id:
        :generation_id:
        :train_fn:
        :args_dict:
        :run_id:

    Returns:

    """

    def _wrapper_fun(iter):
        """

        Args:
            :iter:

        Returns:

        """

        for i in iter:
            executor_num = i

        experiment_utils._set_ml_id(app_id, run_id)

        tb_hdfs_path = ''
        hdfs_exec_logdir = ''

        t = threading.Thread(target=devices._print_periodic_gpu_utilization)
        if devices.get_num_gpus() > 0:
            t.start()

        global local_logdir_bool

        try:
            #Arguments
            if args_dict:
                param_string, params, args = experiment_utils.build_parameters(train_fn, executor_num, args_dict)
                val = _get_return_file(param_string, app_id, generation_id, run_id)
                hdfs_exec_logdir, hdfs_appid_logdir = experiment_utils._create_experiment_subdirectories(app_id, run_id, param_string, 'differential_evolution', sub_type='generation.' + str(generation_id), params=params)
                logfile = experiment_utils._init_logger(hdfs_exec_logdir)
                tb_hdfs_path, tb_pid = tensorboard._register(hdfs_exec_logdir, hdfs_appid_logdir, executor_num, local_logdir=local_logdir_bool)
                print(devices._get_gpu_info())
                print('-------------------------------------------------------')
                print('Started running task ' + param_string)
                if val is not None:
                    val = json.loads(val)
                task_start = time.time()
                if val is None:
                    val = train_fn(*args)
                task_end = time.time()
                time_str = 'Finished task ' + param_string + ' - took ' + experiment_utils._time_diff(task_start, task_end)
                print(time_str)
                experiment_utils._handle_return(val, hdfs_exec_logdir, opt_key, logfile)
                print('Returning metric ' + str(val))
                print('-------------------------------------------------------')
        except:
            raise
        finally:
            experiment_utils._cleanup(tensorboard, t)

    return _wrapper_fun

def _get_return_file(param_string, app_id, generation_id, run_id):
    """

    Args:
        :param_string:
        :app_id:
        :generation_id:
        :run_id:

    Returns:

    """
    handle = hdfs.get()
    for i in range(generation_id):
        possible_result_path = experiment_utils._get_experiments_dir() + '/' + app_id + '_' \
                               + str(run_id) + '/generation.' + str(i) + '/' + param_string + '/.outputs.json'
        if handle.exists(possible_result_path):
            return_file_contents = hdfs.load(possible_result_path)
            return return_file_contents

    return None

def _get_best(root_logdir, direction):

    min_val = sys.float_info.max
    min_logdir = None

    max_val = sys.float_info.min
    max_logdir = None

    generation_folders = hdfs.ls(root_logdir)
    generation_folders.sort()

    for generation in generation_folders:
        for individual in hdfs.ls(generation):
            invidual_files = hdfs.ls(individual, recursive=True)
            for file in invidual_files:
                if file.endswith("/.metric"):
                    val = hdfs.load(file)
                    val = float(val)

                    if val > max_val:
                        max_val = val
                        max_logdir = file[:-8]

                    if val < min_val:
                        min_val = val
                        min_logdir = file[:-8]



    if direction.upper() == Direction.MAX:
        return_dict = {}
        with hdfs.open_file(max_logdir + '/.outputs.json', flags="r") as fi:
            return_dict = json.loads(fi.read())
            fi.close()
        return max_logdir, return_dict
    else:
        return_dict = {}
        with hdfs.open_file(min_logdir + '/.outputs.json', flags="r") as fi:
            return_dict = json.loads(fi.read())
            fi.close()
        return min_logdir, return_dict


"""
A feature store client. This module exposes an API for interacting with feature stores in Hopsworks.
"""

import warnings


def fs_formatwarning(message, category, filename, lineno, line=None):
    return "{}:{}: {}: {}\n".format(filename, lineno, category.__name__, message)


class FeatureStoreDeprecationWarning(Warning):
    """A Warning to be raised when the featurestore module is imported."""
    pass

warnings.formatwarning = fs_formatwarning
warnings.simplefilter("always", FeatureStoreDeprecationWarning)

raise FeatureStoreDeprecationWarning("The `featurestore` module was deprecated with the"
    " introduction of the new `HSFS` client libraries to interact with the Hopsworks "
    " Feature Store. All functionality has been removed from this module.")



"""
A module for getting TLS certificates in YARN containers, used for setting up Kafka inside Jobs/Notebooks on Hops.
"""

import string
import base64
import textwrap
from hops import constants
import os
from pathlib import Path

try:
    import jks
except:
    pass


def _get_key_store_path():
    """
    Get keystore path 

    Returns:
        keystore path  
    """
    k_certificate = Path(constants.SSL_CONFIG.K_CERTIFICATE_CONFIG)
    if k_certificate.exists():
        return k_certificate
    else: 
        username = os.environ['HADOOP_USER_NAME']
        material_directory = Path(os.environ['MATERIAL_DIRECTORY'])
        return material_directory.joinpath(username + constants.SSL_CONFIG.KEYSTORE_SUFFIX)


def get_key_store():
    return str(_get_key_store_path())


def _get_trust_store_path():
    """
    Get truststore location

    Returns:
         truststore location 
    """
    t_certificate = Path(constants.SSL_CONFIG.T_CERTIFICATE_CONFIG)
    if t_certificate.exists():
        return str(t_certificate)
    else: 
        username = os.environ['HADOOP_USER_NAME']
        material_directory = Path(os.environ['MATERIAL_DIRECTORY'])
        return str(material_directory.joinpath(username + constants.SSL_CONFIG.TRUSTSTORE_SUFFIX))


def get_trust_store():
    return str(_get_trust_store_path())


def _get_cert_pw():
    """
    Get keystore password from local container

    Returns:
        Certificate password
    """
    pwd_path = Path(constants.SSL_CONFIG.CRYPTO_MATERIAL_PASSWORD)
    if not pwd_path.exists():
        username = os.environ['HADOOP_USER_NAME']
        material_directory = Path(os.environ['MATERIAL_DIRECTORY'])
        pwd_path = material_directory.joinpath(username + constants.SSL_CONFIG.PASSWORD_SUFFIX)

    with pwd_path.open() as f:
        return f.read()


def get_key_store_cert():
    """
    Get keystore certificate from local container

    Returns:
        Certificate password
    """
    cert_path = _get_key_store_path() 

    if not cert_path.exists():
        raise AssertionError('k_certificate is not present in directory: {}'.format(str(cert_path)))

    # read as bytes, don't try to use utf-8 encoding
    with cert_path.open("rb") as f:
        key_store_cert = f.read()
        key_store_cert = base64.b64encode(key_store_cert)

    return key_store_cert


def get_key_store_pwd():
    """
    Get keystore password

    Returns:
         keystore password
    """
    return _get_cert_pw()


def get_trust_store_pwd():
    """
    Get truststore password

    Returns:
         truststore password
    """
    return _get_cert_pw()


def _bytes_to_pem_str(der_bytes, pem_type):
    """
    Utility function for creating PEM files

    Args:
        der_bytes: DER encoded bytes
        pem_type: type of PEM, e.g Certificate, Private key, or RSA private key

    Returns:
        PEM String for a DER-encoded certificate or private key
    """
    pem_str = ""
    pem_str = pem_str + "-----BEGIN {}-----".format(pem_type) + "\n"
    pem_str = pem_str + "\r\n".join(textwrap.wrap(base64.b64encode(der_bytes).decode('ascii'), 64)) + "\n"
    pem_str = pem_str + "-----END {}-----".format(pem_type) + "\n"
    return pem_str


def _convert_jks_to_pem(jks_path, keystore_pw):
    """
    Converts a keystore JKS that contains client private key,
     client certificate and CA certificate that was used to
     sign the certificate, to three PEM-format strings.

    Args:
    :jks_path: path to the JKS file
    :pw: password for decrypting the JKS file

    Returns:
         strings: (client_cert, client_key, ca_cert)
    """
    # load the keystore and decrypt it with password
    ks = jks.KeyStore.load(jks_path, keystore_pw, try_decrypt_keys=True)
    private_keys_certs = ""
    private_keys = ""
    ca_certs = ""

    # Convert private keys and their certificates into PEM format and append to string
    for alias, pk in ks.private_keys.items():
        if pk.algorithm_oid == jks.util.RSA_ENCRYPTION_OID:
            private_keys = private_keys + _bytes_to_pem_str(pk.pkey, "RSA PRIVATE KEY")
        else:
            private_keys = private_keys + _bytes_to_pem_str(pk.pkey_pkcs8, "PRIVATE KEY")
        for c in pk.cert_chain:
            # c[0] contains type of cert, i.e X.509
            private_keys_certs = private_keys_certs + _bytes_to_pem_str(c[1], "CERTIFICATE")

    # Convert CA Certificates into PEM format and append to string
    for alias, c in ks.certs.items():
        ca_certs = ca_certs + _bytes_to_pem_str(c.cert, "CERTIFICATE")
    return private_keys_certs, private_keys, ca_certs


def _write_pem(jks_key_store_path, jks_trust_store_path, keystore_pw, client_key_cert_path, client_key_path, ca_cert_path):
    """
    Converts the JKS keystore, JKS truststore, and the root ca.pem
    client certificate, client key, and ca certificate

    Args:
    :jks_key_store_path: path to the JKS keystore
    :jks_trust_store_path: path to the JKS truststore
    :keystore_pw: path to file with passphrase for the keystores
    :client_key_cert_path: path to write the client's certificate for its private key in PEM format
    :client_key_path: path to write the client's private key in PEM format
    :ca_cert_path: path to write the chain of CA certificates required to validate certificates
    """
    keystore_key_cert, keystore_key, keystore_ca_cert = _convert_jks_to_pem(jks_key_store_path, keystore_pw)
    truststore_key_cert, truststore_key, truststore_ca_cert = _convert_jks_to_pem(jks_trust_store_path, keystore_pw)
    with client_key_cert_path.open("w") as f:
        f.write(keystore_key_cert)
    with client_key_path.open("w") as f:
        f.write(keystore_key)
    with ca_cert_path.open("w") as f:
        f.write(keystore_ca_cert + truststore_ca_cert)


def get_client_certificate_location():
    """
    Get location of client certificate (PEM format) for the private key signed by trusted CA
    used for 2-way TLS authentication, for example with Kafka cluster

    Returns:
        string path to client certificate in PEM format
    """
    certificate_path = Path(constants.SSL_CONFIG.PEM_CLIENT_CERTIFICATE_CONFIG)
    if not certificate_path.exists(): 
        _write_pems()
    return str(certificate_path)


def get_client_key_location():
    """
    Get location of client private key (PEM format)
    used for for 2-way TLS authentication, for example with Kafka cluster

    Returns:
        string path to client private key in PEM format
    """
    # Convert JKS to PEMs if they don't exists already
    key_path = Path(constants.SSL_CONFIG.PEM_CLIENT_KEY_CONFIG)
    if not key_path.exists():
        _write_pems()
    return str(key_path)


def get_ca_chain_location():
    """
    Get location of chain of CA certificates (PEM format) that are required to validate the
    private key certificate of the client
    used for 2-way TLS authentication, for example with Kafka cluster

    Returns:
         string path to ca chain of certificate
    """
    ca_chain_path = Path(constants.SSL_CONFIG.PEM_CA_CHAIN_CERTIFICATE_CONFIG) 
    if not ca_chain_path.exists():
        _write_pems()
    return str(ca_chain_path)


def _write_pems():
    """
    Converts JKS keystore file into PEM to be compatible with Python libraries
    """
    t_jks_path = get_trust_store() 
    k_jks_path = get_key_store() 

    client_certificate_path = Path(constants.SSL_CONFIG.PEM_CLIENT_CERTIFICATE_CONFIG)
    client_key_path = Path(constants.SSL_CONFIG.PEM_CLIENT_KEY_CONFIG)
    ca_chain_path = Path(constants.SSL_CONFIG.PEM_CA_CHAIN_CERTIFICATE_CONFIG)

    _write_pem(k_jks_path, t_jks_path, get_key_store_pwd(), client_certificate_path, client_key_path, ca_chain_path)

